import os
import unittest

import numpy as np
from rubin_scheduler.data import get_data_dir

import rubin_sim.phot_utils.signaltonoise as snr
from rubin_sim.phot_utils import Bandpass, PhotometricParameters, Sed
from rubin_sim.phot_utils.utils import set_m5


class TestSNRmethods(unittest.TestCase):
    def setUp(self):
        star_name = os.path.join(get_data_dir(), "tests", "cartoonSedTestData", "starSed")
        star_name = os.path.join(star_name, "kurucz", "km20_5750.fits_g40_5790.gz")
        self.star_sed = Sed()
        self.star_sed.read_sed_flambda(star_name)
        imsimband = Bandpass()
        imsimband.imsim_bandpass()
        f_norm = self.star_sed.calc_flux_norm(22.0, imsimband)
        self.star_sed.multiply_flux_norm(f_norm)

        hardware_dir = os.path.join(get_data_dir(), "throughputs", "baseline")
        component_list = [
            "detector.dat",
            "m1.dat",
            "m2.dat",
            "m3.dat",
            "lens1.dat",
            "lens2.dat",
            "lens3.dat",
        ]
        self.sky_sed = Sed()
        self.sky_sed.read_sed_flambda(os.path.join(hardware_dir, "darksky.dat"))

        total_name_list = [
            "total_u.dat",
            "total_g.dat",
            "total_r.dat",
            "total_i.dat",
            "total_z.dat",
            "total_y.dat",
        ]

        self.bp_list = []
        self.hardware_list = []
        for name in total_name_list:
            dummy = Bandpass()
            dummy.read_throughput(os.path.join(hardware_dir, name))
            self.bp_list.append(dummy)

            dummy = Bandpass()
            hardware_name_list = [os.path.join(hardware_dir, name)]
            for component in component_list:
                hardware_name_list.append(os.path.join(hardware_dir, component))
            dummy.read_throughput_list(hardware_name_list)
            self.hardware_list.append(dummy)

        self.filter_name_list = ["u", "g", "r", "i", "z", "y"]

        self.seeing_defaults = {
            "u": 0.92,
            "g": 0.87,
            "r": 0.83,
            "i": 0.80,
            "z": 0.78,
            "y": 0.76,
        }

    def test_mag_error(self):
        """
        Make sure that calc_mag_error_sed and calc_mag_error_m5
        agree to within 0.001
        """
        phot_params = PhotometricParameters()

        # create a cartoon spectrum to test on
        spectrum = Sed()
        spectrum.set_flat_sed()
        spectrum.multiply_flux_norm(1.0e-9)

        # find the magnitudes of that spectrum in our bandpasses
        mag_list = []
        for total in self.bp_list:
            mag_list.append(spectrum.calc_mag(total))
        mag_list = np.array(mag_list)

        # try for different normalizations of the skySED
        for f_norm in np.arange(1.0, 5.0, 1.0):
            self.sky_sed.multiply_flux_norm(f_norm)

            for total, hardware, filterName, mm in zip(
                self.bp_list, self.hardware_list, self.filter_name_list, mag_list
            ):
                fwhm_eff = self.seeing_defaults[filterName]

                m5 = snr.calc_m5(self.sky_sed, total, hardware, phot_params, fwhm_eff=fwhm_eff)

                sigma_sed = snr.calc_mag_error_sed(
                    spectrum,
                    total,
                    self.sky_sed,
                    hardware,
                    phot_params,
                    fwhm_eff=fwhm_eff,
                )

                sigma_m5, gamma = snr.calc_mag_error_m5(mm, total, m5, phot_params)

                self.assertAlmostEqual(sigma_m5, sigma_sed, 3)

    def test_verbose_snr(self):
        """
        Make sure that calc_snr_sed has everything
        it needs to run in verbose mode
        """
        phot_params = PhotometricParameters()

        # create a cartoon spectrum to test on
        spectrum = Sed()
        spectrum.set_flat_sed()
        spectrum.multiply_flux_norm(1.0e-9)

        snr.calc_snr_sed(
            spectrum,
            self.bp_list[0],
            self.sky_sed,
            self.hardware_list[0],
            phot_params,
            fwhm_eff=0.7,
            verbose=True,
        )

    def test_signal_to_noise(self):
        """
        Test that calc_snr_m5 and calc_snr_sed give similar results
        """
        phot_params = PhotometricParameters()

        m5 = []
        for i in range(len(self.hardware_list)):
            m5.append(
                snr.calc_m5(
                    self.sky_sed,
                    self.bp_list[i],
                    self.hardware_list[i],
                    phot_params,
                    fwhm_eff=self.seeing_defaults[self.filter_name_list[i]],
                )
            )

        sed_dir = os.path.join(get_data_dir(), "tests", "cartoonSedTestData/starSed/")
        sed_dir = os.path.join(sed_dir, "kurucz")
        file_name_list = os.listdir(sed_dir)

        rng = np.random.RandomState(42)
        offset = rng.random_sample(len(file_name_list)) * 2.0

        for ix, name in enumerate(file_name_list):
            if ix > 100:
                break
            spectrum = Sed()
            spectrum.read_sed_flambda(os.path.join(sed_dir, name))
            ff = spectrum.calc_flux_norm(m5[2] - offset[ix], self.bp_list[2])
            spectrum.multiply_flux_norm(ff)
            for i in range(len(self.bp_list)):
                control_snr = snr.calc_snr_sed(
                    spectrum,
                    self.bp_list[i],
                    self.sky_sed,
                    self.hardware_list[i],
                    phot_params,
                    self.seeing_defaults[self.filter_name_list[i]],
                )

                mag = spectrum.calc_mag(self.bp_list[i])

                test_snr, gamma = snr.calc_snr_m5(mag, self.bp_list[i], m5[i], phot_params)
                self.assertLess((test_snr - control_snr) / control_snr, 0.001)

    def test_systematic_uncertainty(self):
        """
        Test that systematic uncertainty is added correctly.
        """
        sigma_sys = 0.002
        m5_list = [23.5, 24.3, 22.1, 20.0, 19.5, 21.7]
        phot_params = PhotometricParameters(sigma_sys=sigma_sys)

        magnitude_list = []
        for bp in self.bp_list:
            mag = self.star_sed.calc_mag(bp)
            magnitude_list.append(mag)

        for bp, hardware, filterName, mm, m5 in zip(
            self.bp_list,
            self.hardware_list,
            self.filter_name_list,
            magnitude_list,
            m5_list,
        ):
            sky_dummy = Sed()
            sky_dummy.read_sed_flambda(os.path.join(get_data_dir(), "throughputs", "baseline", "darksky.dat"))

            normalized_sky_dummy = set_m5(
                m5,
                sky_dummy,
                bp,
                hardware,
                fwhm_eff=self.seeing_defaults[filterName],
                phot_params=phot_params,
            )

            sigma, gamma = snr.calc_mag_error_m5(mm, bp, m5, phot_params)

            snrat = snr.calc_snr_sed(
                self.star_sed,
                bp,
                normalized_sky_dummy,
                hardware,
                fwhm_eff=self.seeing_defaults[filterName],
                phot_params=PhotometricParameters(),
            )

            test_snr, gamma = snr.calc_snr_m5(mm, bp, m5, phot_params=PhotometricParameters(sigma_sys=0.0))

            self.assertAlmostEqual(
                snrat,
                test_snr,
                10,
                msg="failed on calc_snr_m5 test %e != %e " % (snrat, test_snr),
            )

            control = np.sqrt(np.power(snr.mag_error_from_snr(test_snr), 2) + np.power(sigma_sys, 2))

            msg = "%e is not %e; failed" % (sigma, control)

            self.assertAlmostEqual(sigma, control, 10, msg=msg)

    def test_no_systematic_uncertainty(self):
        """
        Test that systematic uncertainty is handled correctly
        when set to None.
        """
        m5_list = [23.5, 24.3, 22.1, 20.0, 19.5, 21.7]
        phot_params = PhotometricParameters(sigma_sys=0.0)

        magnitude_list = []
        for bp in self.bp_list:
            mag = self.star_sed.calc_mag(bp)
            magnitude_list.append(mag)

        for bp, hardware, filterName, mm, m5 in zip(
            self.bp_list,
            self.hardware_list,
            self.filter_name_list,
            magnitude_list,
            m5_list,
        ):
            sky_dummy = Sed()
            sky_dummy.read_sed_flambda(os.path.join(get_data_dir(), "throughputs", "baseline", "darksky.dat"))

            normalized_sky_dummy = set_m5(
                m5,
                sky_dummy,
                bp,
                hardware,
                fwhm_eff=self.seeing_defaults[filterName],
                phot_params=phot_params,
            )

            sigma, gamma = snr.calc_mag_error_m5(mm, bp, m5, phot_params)

            snrat = snr.calc_snr_sed(
                self.star_sed,
                bp,
                normalized_sky_dummy,
                hardware,
                fwhm_eff=self.seeing_defaults[filterName],
                phot_params=PhotometricParameters(),
            )

            test_snr, gamma = snr.calc_snr_m5(mm, bp, m5, phot_params=PhotometricParameters(sigma_sys=0.0))

            self.assertAlmostEqual(
                snrat,
                test_snr,
                10,
                msg="failed on calc_snr_m5 test %e != %e " % (snrat, test_snr),
            )

            control = snr.mag_error_from_snr(test_snr)

            msg = "%e is not %e; failed" % (sigma, control)

            self.assertAlmostEqual(sigma, control, 10, msg=msg)

    def test_fwh_mconversions(self):
        fwhm_eff = 0.8
        fwh_mgeom = snr.fwhm_eff2_fwhm_geom(fwhm_eff)
        self.assertEqual(fwh_mgeom, (0.822 * fwhm_eff + 0.052))
        fwh_mgeom = 0.8
        fwhm_eff = snr.fwhm_geom2_fwhm_eff(fwh_mgeom)
        self.assertEqual(fwhm_eff, (fwh_mgeom - 0.052) / 0.822)

    def test_astrometric_error(self):
        fwhm_geom = 0.7
        m5 = 24.5
        # For bright objects, error should be systematic floor
        mag = 10
        astrometric_err = snr.calc_astrometric_error(
            mag, m5, fwhm_geom=fwhm_geom, nvisit=1, systematic_floor=10
        )
        self.assertAlmostEqual(astrometric_err, 10, 3)
        # Even if you increase the number of visits,
        # the systemic floor doesn't change
        astrometric_err = snr.calc_astrometric_error(mag, m5, fwhm_geom=fwhm_geom, nvisit=100)
        self.assertAlmostEqual(astrometric_err, 10, 3)
        # For a single visit, fainter source, larger error and nvisits matters
        mag = 24.5
        astrometric_err1 = snr.calc_astrometric_error(
            mag, m5, fwhm_geom=fwhm_geom, nvisit=1, systematic_floor=10
        )
        astrometric_err100 = snr.calc_astrometric_error(
            mag, m5, fwhm_geom=fwhm_geom, nvisit=100, systematic_floor=10
        )
        self.assertGreater(astrometric_err1, astrometric_err100)
        self.assertAlmostEqual(astrometric_err1, 140.357, 3)

    def test_snr_arr(self):
        """
        Test that calc_snr_m5 works on numpy arrays of magnitudes
        """
        rng = np.random.RandomState(17)
        mag_list = rng.random_sample(100) * 5.0 + 15.0

        phot_params = PhotometricParameters()
        bp = self.bp_list[0]
        m5 = 24.0
        control_list = []
        for mm in mag_list:
            ratio, gamma = snr.calc_snr_m5(mm, bp, m5, phot_params)
            control_list.append(ratio)
        control_list = np.array(control_list)

        test_list, gamma = snr.calc_snr_m5(mag_list, bp, m5, phot_params)

        np.testing.assert_array_equal(control_list, test_list)

    def test_error_arr(self):
        """
        Test that calc_mag_error_m5 works on numpy arrays of magnitudes
        """
        rng = np.random.RandomState(17)
        mag_list = rng.random_sample(100) * 5.0 + 15.0

        phot_params = PhotometricParameters()
        bp = self.bp_list[0]
        m5 = 24.0
        control_list = []
        for mm in mag_list:
            sig, gamma = snr.calc_mag_error_m5(mm, bp, m5, phot_params)
            control_list.append(sig)
        control_list = np.array(control_list)

        test_list, gamma = snr.calc_mag_error_m5(mag_list, bp, m5, phot_params)

        np.testing.assert_array_equal(control_list, test_list)


if __name__ == "__main__":
    unittest.main()



import os
import unittest

import numpy as np
from rubin_scheduler.data import get_data_dir

import rubin_sim.skybrightness as sb
from rubin_sim.phot_utils import Bandpass


class TestSkyModel(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # initalize the class with empty models
        cls._sm_mags = None
        cls._sm_mags2 = None
        cls._sm_spec = None
        cls._sm_spec2 = None

    @classmethod
    def tearDownClass(cls):
        del cls._sm_mags
        del cls._sm_mags2
        del cls._sm_spec
        del cls._sm_spec2

    @property
    def sm_mags(self):
        cls = type(self)
        if cls._sm_mags is None:
            cls._sm_mags = sb.SkyModel(mags=True)
        return cls._sm_mags

    @property
    def sm_mags2(self):
        cls = type(self)
        if cls._sm_mags2 is None:
            cls._sm_mags2 = sb.SkyModel(mags=True)
        return cls._sm_mags2

    @property
    def sm_spec(self):
        cls = type(self)
        if cls._sm_spec is None:
            cls._sm_spec = sb.SkyModel(mags=False)
        return cls._sm_spec

    @property
    def sm_spec2(self):
        cls = type(self)
        if cls._sm_spec2 is None:
            cls._sm_spec2 = sb.SkyModel(mags=False)
        return cls._sm_spec2

    def testmerged_comp(self):
        """
        Test that the 3 components that have been merged return the
        same result if they are computed independently
        """

        sky1 = sb.SkyModel(
            twilight=False,
            zodiacal=False,
            moon=False,
            lower_atm=False,
            upper_atm=False,
            airglow=False,
            scattered_star=False,
            merged_spec=True,
        )
        sky1.set_ra_dec_mjd([36.0], [-68.0], 49353.18, degrees=True)

        sky2 = sb.SkyModel(
            twilight=False,
            zodiacal=False,
            moon=False,
            lower_atm=True,
            upper_atm=True,
            airglow=False,
            scattered_star=True,
            merged_spec=False,
        )
        sky2.set_ra_dec_mjd([36.0], [-68.0], 49353.18, degrees=True)

        dummy, spec1 = sky1.return_wave_spec()
        dummy, spec2 = sky2.return_wave_spec()

        np.testing.assert_almost_equal(spec1, spec2)

        # and then check for the mags
        sky1 = sb.SkyModel(
            twilight=False,
            zodiacal=False,
            moon=False,
            lower_atm=False,
            upper_atm=False,
            airglow=False,
            scattered_star=False,
            merged_spec=True,
            mags=True,
        )
        sky1.set_ra_dec_mjd([36.0], [-68.0], 49353.18, degrees=True)

        sky2 = sb.SkyModel(
            twilight=False,
            zodiacal=False,
            moon=False,
            lower_atm=True,
            upper_atm=True,
            airglow=False,
            scattered_star=True,
            merged_spec=False,
            mags=True,
        )
        sky2.set_ra_dec_mjd([36.0], [-68.0], 49353.18, degrees=True)

        m1 = sky1.return_mags()
        m2 = sky2.return_mags()
        for key in m1:
            np.testing.assert_almost_equal(m1[key], m2[key], decimal=2)

    def test_setups(self):
        """
        Check that things are the same if the model is set up with
        radecmjd or all the parameters independently
        """

        sm1 = self.sm_spec
        sm1.set_ra_dec_mjd([36.0], [-68.0], 49353.18, degrees=True)

        sm2 = self.sm_spec2
        sm2.set_params(
            azs=sm1.azs,
            alts=sm1.alts,
            moon_phase=sm1.moon_phase,
            moon_alt=sm1.moon_alt,
            moon_az=sm1.moon_az,
            sun_alt=sm1.sun_alt,
            sun_az=sm1.sun_az,
            sun_eclip_lon=sm1.sun_eclip_lon,
            eclip_lon=sm1.eclip_lon,
            eclip_lat=sm1.eclip_lat,
            solar_flux=sm1.solar_flux,
            degrees=False,
        )

        dummy, spec1 = sm1.return_wave_spec()
        dummy, spec2 = sm2.return_wave_spec()

        np.testing.assert_array_equal(spec1, spec2)

        # Check that the degrees kwarg works
        sm2.set_params(
            azs=np.degrees(sm1.azs),
            alts=np.degrees(sm1.alts),
            moon_phase=sm1.moon_phase,
            moon_alt=np.degrees(sm1.moon_alt),
            moon_az=np.degrees(sm1.moon_az),
            sun_alt=np.degrees(sm1.sun_alt),
            sun_az=np.degrees(sm1.sun_az),
            sun_eclip_lon=np.degrees(sm1.sun_eclip_lon),
            eclip_lon=np.degrees(sm1.eclip_lon),
            eclip_lat=np.degrees(sm1.eclip_lat),
            solar_flux=sm1.solar_flux,
            degrees=True,
        )

        at_list = [
            "azs",
            "alts",
            "moon_phase",
            "moon_alt",
            "moon_az",
            "sun_alt",
            "sun_az",
            "sun_eclip_lon",
            "eclip_lon",
            "eclip_lat",
            "solar_flux",
        ]

        # Check each attribute that should match
        for attr in at_list:
            np.testing.assert_allclose(getattr(sm1, attr), getattr(sm2, attr))

        # Check the interpolation points
        for name in sm1.points.dtype.names:
            np.testing.assert_allclose(sm1.points[name], sm2.points[name])

        # Check the final output spectra
        np.testing.assert_allclose(sm1.spec, sm2.spec)

    def test_mags(self):
        """
        Test that the interpolated mags are similar to mags computed from interpolated spectra
        """

        through_path = os.path.join(get_data_dir(), "throughputs", "baseline")
        filters = ["u", "g", "r", "i", "z", "y"]

        bps = {}
        for filter_name in filters:
            bp = np.loadtxt(
                os.path.join(through_path, "total_%s.dat" % filter_name),
                dtype=list(zip(["wave", "trans"], [float] * 2)),
            )
            lsst_bp = Bandpass()
            lsst_bp.set_bandpass(bp["wave"], bp["trans"])
            bps[filter_name] = lsst_bp

        sm1 = self.sm_spec
        sm1.set_ra_dec_mjd([36.0], [-68.0], 49353.18, degrees=True)
        mags1 = sm1.return_mags(bandpasses=bps)

        sm2 = self.sm_mags
        sm2.set_ra_dec_mjd([36.0], [-68.0], 49353.18, degrees=True)
        mag2 = sm2.return_mags()

        # Let's set the tolerance for matching the throughputs to be 0.001
        # This allows for small changes in the filter throughputs without requiring recomputation of
        # sims_skybrighntess_pre, while still requiring a reasonable match against the measured sky
        for i, filtername in enumerate(filters):
            np.testing.assert_allclose(mags1[filtername], mag2[filtername], rtol=5e-3)

    def test_get_computed(self):
        """
        Make sure we can recover computed values.
        """

        sm = self.sm_mags
        sm.set_ra_dec_mjd([36.0, 36.0], [-68.0, -70.0], 49353.18, degrees=True)
        val_dict = sm.get_computed_vals()

        attr_to_check = [
            "ra",
            "dec",
            "alts",
            "azs",
            "airmass",
            "solar_flux",
            "moon_phase",
            "moon_az",
            "moon_alt",
            "sun_alt",
            "sun_az",
            "az_rel_sun",
            "moon_sun_sep",
            "az_rel_moon",
            "eclip_lon",
            "eclip_lat",
            "moon_ra",
            "moon_dec",
            "sun_ra",
            "sun_dec",
            "sun_eclip_lon",
        ]

        for attr in attr_to_check:
            assert attr in val_dict
            if np.size(val_dict[attr]) > 1:
                np.testing.assert_array_equal(getattr(sm, attr), val_dict[attr])
            else:
                self.assertEqual(getattr(sm, attr), val_dict[attr])

        # Check that things that should be radians are in radian range
        rad_list = [
            "ra",
            "azs",
            "moon_az",
            "sun_az",
            "az_rel_sun",
            "az_rel_moon",
            "eclip_lon",
            "moon_ra",
            "sun_ra",
            "sun_eclip_lon",
        ]

        for attr in rad_list:
            if val_dict[attr] is not None:
                assert np.min(val_dict[attr]) >= 0
                assert np.max(val_dict[attr]) <= 2.0 * np.pi

        # Radians in negative to positive pi range
        rad_list = [
            "moon_alt",
            "sun_alt",
            "alts",
            "dec",
            "moon_dec",
            "sun_dec",
            "eclip_lat",
        ]
        for attr in rad_list:
            if val_dict[attr] is not None:
                assert np.min(val_dict[attr]) >= -np.pi
                assert np.max(val_dict[attr]) <= np.pi

    def test90_deg(self):
        """
        Make sure we can look all the way to 90 degree altitude.
        """
        mjd = 56973.268218
        sm = self.sm_mags
        sm.set_ra_dec_mjd(0.0, 90.0, mjd, degrees=True, az_alt=True)
        mags = sm.return_mags()
        for key in mags:
            assert True not in np.isnan(mags[key])
        assert True not in np.isnan(sm.spec)

    def test_fewer_mags(self):
        """
        Test that can call and only interpolate a few magnitudes.
        """
        mjd = 56973.268218
        sm = self.sm_mags
        sm.set_ra_dec_mjd(0.0, 90.0, mjd, degrees=True, az_alt=True)
        all_mags = sm.return_mags()

        filter_names = ["u", "g", "r", "i", "z", "y"]
        for filter_name in filter_names:
            sm.set_ra_dec_mjd(0.0, 90.0, mjd, degrees=True, az_alt=True, filter_names=[filter_name])
            one_mag = sm.return_mags()
            self.assertEqual(all_mags[filter_name], one_mag[filter_name])

        # Test that I can do subset of mags
        subset = ["u", "r", "y"]
        sm.set_ra_dec_mjd(0.0, 90.0, mjd, degrees=True, az_alt=True, filter_names=subset)
        sub_mags = sm.return_mags()
        for filter_name in subset:
            self.assertEqual(all_mags[filter_name], sub_mags[filter_name])

    def test_set_ra_dec_alt_az_mjd(self):
        """
        Make sure sending in self-computed alt, az works
        """
        sm1 = self.sm_mags
        sm2 = self.sm_mags2
        ra = np.array([0.0, 0.0, 0.0])
        dec = np.array([-0.1, -0.2, -0.3])
        mjd = 5900
        sm1.set_ra_dec_mjd(ra, dec, mjd)
        m1 = sm1.return_mags()
        sm2.set_ra_dec_alt_az_mjd(ra, dec, sm1.alts, sm1.azs, mjd)
        m2 = sm1.return_mags()

        attr_list = ["ra", "dec", "alts", "azs"]
        for attr in attr_list:
            np.testing.assert_equal(getattr(sm1, attr), getattr(sm2, attr))

        for key in m1:
            np.testing.assert_allclose(m1[key], m2[key], rtol=1e-6)


if __name__ == "__main__":
    unittest.main()



__all__ = ("PeriodDeviationMetric",)

import numpy as np
from scipy.signal import lombscargle

from rubin_sim.maf.metrics import BaseMetric

# Example of a *very* simple variabiilty metric
# krughoff@uw.edu, ebellm, ljones


def find_period_ls(times, mags, minperiod=2.0, maxperiod=35.0, nbinmax=10**5, verbose=False):
    """Find the period of a lightcurve using scipy's lombscargle method.
    The parameters used here imply magnitudes but there is no reason this would not work if fluxes are passed.

    :param times: A list of times for the given observations
    :param mags: A list of magnitudes for the object at the given times
    :param minperiod: Minimum period to search
    :param maxperiod: Maximum period to search
    :param nbinmax: Maximum number of frequency bins to use in the search
    :returns: Period in the same units as used in times.  This is simply
              the max value in the Lomb-Scargle periodogram
    """
    if minperiod < 0:
        minperiod = 0.01
    nbins = int((times.max() - times.min()) / minperiod * 1000)
    if nbins > nbinmax:
        if verbose:
            print("lowered nbins")
        nbins = nbinmax

    # Recenter the magnitude measurements about zero
    dmags = mags - np.median(mags)
    # Create frequency bins
    f = np.linspace(1.0 / maxperiod, 1.0 / minperiod, nbins)

    # Calculate periodogram
    pgram = lombscargle(times, dmags, f)

    idx = np.argmax(pgram)
    # Return period of the bin with the max value in the periodogram
    return 1.0 / f[idx]


class PeriodDeviationMetric(BaseMetric):
    """Measure the percentage deviation of recovered periods for pure sine wave variability (in magnitude)."""

    def __init__(
        self,
        col="observationStartMJD",
        period_min=3.0,
        period_max=35.0,
        n_periods=5,
        mean_mag=21.0,
        amplitude=1.0,
        metric_name="Period Deviation",
        period_check=None,
        **kwargs,
    ):
        """
        Construct an instance of a PeriodDeviationMetric class

        :param col: Name of the column to use for the observation times, commonly 'expMJD'
        :param period_min: Minimum period to test (days)
        :param period_max: Maximimum period to test (days)
        :param period_check: Period to use in the reduce function (days)
        :param mean_mag: Mean value of the lightcurve
        :param amplitude: Amplitude of the variation (mags)
        """
        self.period_min = period_min
        self.period_max = period_max
        self.period_check = period_check
        self.guess_p_min = np.min([self.period_min * 0.8, self.period_min - 1])
        self.guess_p_max = np.max([self.period_max * 1.20, self.period_max + 1])
        self.n_periods = n_periods
        self.mean_mag = mean_mag
        self.amplitude = amplitude
        super(PeriodDeviationMetric, self).__init__(col, metric_name=metric_name, **kwargs)

    def run(self, data_slice, slice_point=None):
        """
        Run the PeriodDeviationMetric
        :param data_slice : Data for this slice.
        :param slice_point: Metadata for the slice. (optional)
        :return: The error in the period estimated from a Lomb-Scargle periodogram
        """

        # Make sure the observation times are sorted
        data = np.sort(data_slice[self.colname])

        # Create 'nPeriods' random periods within range of min to max.
        if self.period_check is not None:
            periods = [self.period_check]
        else:
            periods = self.period_min + np.random.random(self.n_periods) * (self.period_max - self.period_min)
        # Make sure the period we want to check is in there
        periodsdev = np.zeros(np.size(periods), dtype="float")
        for i, period in enumerate(periods):
            omega = 1.0 / period
            # Calculate up the amplitude.
            lc = self.mean_mag + self.amplitude * np.sin(omega * data)
            # Try to recover the period given a window buffered by min of a day or 20% of period value.
            if len(lc) < 3:
                # Too few points to find a period
                return self.badval

            pguess = find_period_ls(data, lc, minperiod=self.guess_p_min, maxperiod=self.guess_p_max)
            periodsdev[i] = (pguess - period) / period

        return {"periods": periods, "periodsdev": periodsdev}

    def reduce_p_dev(self, metric_val):
        """
        At a particular slice_point, return the period deviation for self.period_check.
        If self.period_check is None, just return a random period in the range.
        """
        result = metric_val["periodsdev"][0]
        return result

    def reduce_worst_period(self, metric_val):
        """
        At each slice_point, return the period with the worst period deviation.
        """
        worst_p = np.array(metric_val["periods"])[
            np.where(metric_val["periodsdev"] == metric_val["periodsdev"].max())[0]
        ]
        return worst_p

    def reduce_worst_p_dev(self, metric_val):
        """
        At each slice_point, return the largest period deviation.
        """
        worst_p_dev = np.array(metric_val["periodsdev"])[
            np.where(metric_val["periodsdev"] == metric_val["periodsdev"].max())[0]
        ]
        return worst_p_dev



__all__ = ("artificial_structure_calculation",)

#####################################################################################################
# Purpose: calculate artificial structure, i.e. fluctuations in galaxy counts, resulting from
# imperfect observing strategy (OS). Includes the functionality to account for dust extinction,
# photometric calibration errors (simple ansatz implemented here), individual redshift bins (see
# GalaxyCountsMetricExtended for details), as well as poisson noise in the galaxy counts.
#
# Basic workflow, assuming all the functionalities are used:
#       1. HEALpix slicers are set up for survey strategies.
#       2. Using GalaxyCountMetric_extended, which handles dust extinction and calculates galaxy counts
#          based on redshift-bin-specific powerlaws, galaxy counts are found for each HEALpix pixel.
#       3. The shallow borders are masked (based on user-specified 'pixel radius').
#       4. Photometric calibration errors are calculated.
#       5. The galaxy counts in each pixel are recalculated using galaxy_counts_with_pixel_calibration
#          since the calibration errors modify the upper limit on the integral used to calculate
#          galaxy counts. galaxy_counts_with_pixel_calibration takes in each pixel's modified integration
#          limit individually.
#       6. Poisson noise is added to the galaxy counts.
#       7. Fluctuations in the galaxy counts are calculated.
#
# For each pixel i, the photometric calibration errors are modeled as del_i= k*z_i/sqrt(nObs_i),
# where z_i is the average seeing the pixel minus avgSeeing across map, nObs is the number of observations,
# and k is a constant such that var(del_i)= (0.01)^2 -- 0.01 in accordance with LSST goal for relative
# photometric calibration.
#
# Most of the functionalities can be turned on/off, and plots and data can be saved at various points.
# Bordering masking adds significant run time as does the incorporation of photometric calibration
# errors. See the method descrpition for further details.
#
# Humna Awan: humna.awan@rutgers.edu
#####################################################################################################
__all__ = ("artificial_structure_calculation",)

import os

import healpy as hp
import matplotlib.pyplot as plt
import numpy as np

try:
    from sympy import Symbol
    from sympy.solvers import solve
except ImportError:
    pass
import copy
import datetime
import time

from matplotlib.ticker import FuncFormatter

import rubin_sim.maf
import rubin_sim.maf.db as db
import rubin_sim.maf.maps as maps
import rubin_sim.maf.metric_bundles as metricBundles
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.plots as plots
import rubin_sim.maf.slicers as slicers
import rubin_sim.maf.stackers as mafStackers  # stackers in sims_maf
from rubin_sim.maf.maf_contrib.lss_obs_strategy.constants_for_pipeline import plot_color, power_law_const_a
from rubin_sim.maf.maf_contrib.lss_obs_strategy.galaxy_counts_metric_extended import (
    GalaxyCountsMetricExtended as GalaxyCountsMetric,
)
from rubin_sim.maf.maf_contrib.lss_obs_strategy.galaxy_counts_with_pixel_calibration import (
    galaxy_counts_with_pixel_calibration as GalaxyCounts_0ptErrors,
)
from rubin_sim.maf.maf_contrib.lss_obs_strategy.masking_algorithm_generalized import (
    masking_algorithm_generalized,
)
from rubin_sim.maf.metrics import CountMetric as NumObsMetric


def artificial_structure_calculation(
    path,
    upper_mag_limit,
    dbfile,
    run_name,
    no_dith_only=False,
    best_dith_only=False,
    specified_dith=None,
    nside=128,
    filter_band="i",
    cut_off_year=None,
    redshift_bin="all",
    cfhtls_counts=False,
    normalized_mock_catalog_counts=True,
    include_dust_extinction=True,
    save_raw_num_gal_data=True,
    pixel_radius_for_masking=5,
    save_num_gal_data_after_masking=False,
    include0pt_errors=True,
    print0pt_information=True,
    plot0pt_plots=True,
    show0pt_plots=False,
    save0pt_plots=True,
    save_num_gal_data_after0pt=False,
    add_poisson_noise=True,
    save_delta_n_by_n_data=True,
    save_cls_for_delta_n_by_n=True,
    show_comp_plots=False,
    return_stuff=False,
):
    """
    Calculate artificial structure, i.e. fluctuations in galaxy counts dN/N, resulting due
    to imperfect observing strategy (OS).
    - Creates an output directory for subdirectories containing the specified things to save.
    - Prints out execution time at key steps (after border-masking, incorporating calibration errors, etc.)
    - Returns the metricBundle object containing the calculated dN/N, the output directory name,
    the results_db object, and (if include0pt_errors=True)  calibration errors for each survey strategy.

    Parameters
    ----------
    path : str
        path to the main directory where output directory is to be saved.
    upper_mag_limit : float
        upper limit on magnitude when calculating the galaxy counts.
    dbfile : str
        path to the OpSim output file, e.g. to a copy of enigma_1189
    run_name : str
        run name tag to identify the output of specified OpSim output.
        Since new OpSim outputs have different columns, the run_name for enigma_1189 **must**
        be 'enigma1189'; can be anything for other outputs, e.g. 'minion1016'
    no_dith_only : `bool`
        set to True if only want to consider the undithered survey. Default: False
    best_dith_only : `bool`
        set to True if only want to consider RandomDitherFieldPerVisit.
        Default: False
    specified_dith : str
        specific dither strategy to run; could be a string or a list of strings.
        Default: None
    nside : int
        HEALpix resolution parameter. Default: 128
    filter_band : str
        any one of 'u', 'g', 'r', 'i', 'z', 'y'. Default: 'i'
    cut_off_year : int
        year cut to restrict analysis to only a subset of the survey.
        Must range from 1 to 9, or None for the full survey analysis (10 yrs).
        Default: None
    redshift_bin : str
        options include '0.<z<0.15', '0.15<z<0.37', '0.37<z<0.66, '0.66<z<1.0',
        '1.0<z<1.5', '1.5<z<2.0', '2.0<z<2.5', '2.5<z<3.0','3.0<z<3.5', '3.5<z<4.0',
        'all' for no redshift restriction (i.e. 0.<z<4.0)
        Default: 'all'
    cfhtls_counts : `bool`
        set to True if want to calculate the total galaxy counts from CFHTLS
        powerlaw from LSST Science Book. Must be run with redshift_bin='all'
        Default: False
    normalized_mock_catalog_counts : `bool`
        set to False if  want the raw/un-normalized galaxy
        counts from mock catalogs. Default: True
    include_dust_extinction : `bool`:
        set to include dust extinction when calculating the coadded
        depth. Default: True
    save_raw_num_gal_data : `bool`
        set to True to save num_gal data right away, i.e. before
        0pt error calibration, bordering masking, or poisson noise.
        Default: True
    pixel_radius_for_masking : int
        number of pixels to mask along the shallow border. Default: 5
    save_num_gal_data_after_masking : `bool`
        set to True to save num_gal data after border masking.
        Default: False
    include0pt_errors : `bool`
        set to True to include photometric calibration errors.
        Default: True
    print0pt_information : `bool`
        set to True to print out some statistics (variance, the k-value, etc.)
        of the calibration errors of every dither strategy.
        Default: True
    plot0pt_plots : `bool`
        generate 0pt plots. Default True.
    save_num_gal_data_after0pt : `bool`
        set to True to save num_gal data after border masking and 0pt calibration. Default: False
    add_poisson_noise : `bool`
        set to True to add poisson noise to the galaxy counts after border masking
        and the incorporation of calibration errors. Default: True
    saveNumGalDataAfterPoisson : `bool`
        set to True to save num_gal data right away, after border masking,
        including the calibration errors, and the  poisson noise.
        Default: True
    showDeltaNByNPlots : `bool`
        set to True to show the plots related to the fluctuations in the galaxy
        counts. Will work only when plotDeltaNByN=True. Default: False
    saveDeltaNByNPlots : `bool`
        set to True to save the plots related to the fluctuations in the galaxy
        counts. Will work only when plotDeltaNByN=True. Default: True
    save_delta_n_by_n_data : `bool`
        set to True to save data for the the fluctuations in the galaxy counts.
        Default: True
    save_cls_for_delta_n_by_n : `bool`
        set to True to save the power spectrum data for the the fluctuations in
        the galaxy counts. Default: True
    show_comp_plots : `bool`
        set to True if want to display the comparison plots (only valid if have more
        han one dither strategy); otherwise, the plots will be saved and not displayed.
        Default: False
    return_stuff : `bool`
        set to True to get the metricBundle object, the out_dir, and results_db object.
        Default: False
    """
    start_time = time.time()
    # ------------------------------------------------------------------------
    # set up the metric
    gal_count_metric = GalaxyCountsMetric(
        upper_mag_limit=upper_mag_limit,
        include_dust_extinction=include_dust_extinction,
        redshift_bin=redshift_bin,
        filter_band=filter_band,
        nside=nside,
        cfhtls_counts=cfhtls_counts,
        normalized_mock_catalog_counts=normalized_mock_catalog_counts,
    )
    # OpSim database
    opsdb = db.OpsimDatabase(dbfile)

    # ------------------------------------------------------------------------
    # set up the out_dir name
    zeropt_tag, dust_tag = "", ""
    if include0pt_errors:
        zeropt_tag = "with0ptErrors"
    else:
        zeropt_tag = "no0ptErrors"

    if include_dust_extinction:
        dust_tag = "withDustExtinction"
    else:
        dust_tag = "noDustExtinction"

    if cut_off_year is not None:
        survey_tag = "%syearCut" % (cut_off_year)
    else:
        survey_tag = "fullSurveyPeriod"

    # check to make sure redshift bin is ok.
    allowed_redshift_bins = list(power_law_const_a.keys()) + ["all"]
    if redshift_bin not in allowed_redshift_bins:
        print("ERROR: Invalid redshift bin. Input bin can only be among %s.\n" % (allowed_redshift_bins))
        return
    zbin_tag = redshift_bin
    if redshift_bin == "all":
        zbin_tag = "allRedshiftData"

    poisson_tag = ""
    if add_poisson_noise:
        poisson_tag = "withPoissonNoise"
    else:
        poisson_tag = "noPoissonNoise"

    counts_tag = ""
    if cfhtls_counts:
        counts_tag = "CFHTLSpowerLaw"
    elif normalized_mock_catalog_counts:
        counts_tag = "normalizedGalaxyCounts"
    else:
        counts_tag = "unnormalizedGalaxyCounts"

    out_dir = (
        f"artificialStructure_{poisson_tag}_nside{nside}"
        f"_pixelRadiusFormasking_{pixel_radius_for_masking}_{zeropt_tag}_{dust_tag}_{filter_band}"
        f"_{upper_mag_limit}_{run_name}_{survey_tag}_{zbin_tag}_{counts_tag}_directory"
    )

    print("# out_dir: %s\n" % out_dir)
    if not os.path.exists("%s%s" % (path, out_dir)):
        os.makedirs("%s%s" % (path, out_dir))

    results_dbname = "resultsDb_%s.db" % np.random.randint(100)
    results_db = db.ResultsDb(database=results_dbname, out_dir="%s%s" % (path, out_dir))

    # ------------------------------------------------------------------------
    # set up the sql constraint
    prop_ids, prop_tags = opsdb.fetchPropInfo()
    wfd_where = opsdb.createSQLWhere("WFD", prop_tags)
    ra_dec_in_deg = opsdb.raDecInDeg
    if cut_off_year is not None:
        night_cut_off = (cut_off_year) * 365.25
        sqlconstraint = '%s and night<=%s and filter=="%s"' % (
            wfd_where,
            night_cut_off,
            filter_band,
        )
    else:
        sqlconstraint = '%s and filter=="%s"' % (wfd_where, filter_band)
    print("# sqlconstraint: %s" % sqlconstraint)

    # ------------------------------------------------------------------------
    # create a ReadMe type file to put info in.
    update = "%s\n" % (datetime.datetime.now())
    update += "\nArtificial structure calculation with %s, %s, and %s " % (
        zeropt_tag,
        dust_tag,
        poisson_tag,
    )
    update += "for %s for %s for %s<%s. " % (
        survey_tag,
        zbin_tag,
        filter_band,
        upper_mag_limit,
    )
    update += "\nWith %s and PixelRadiusForMasking: %s.\n" % (
        counts_tag,
        pixel_radius_for_masking,
    )
    update += "\nsqlconstraint: %s" % sqlconstraint
    update += "\nRunning with %s\n" % run_name
    update += "\nout_dir: %s\n" % out_dir
    update += "\nMAF version: %s\n" % rubin_sim.maf.__version__

    # figure out the readme name
    readme_name = "ReadMe"
    readmes = [f for f in os.listdir("%s%s" % (path, out_dir)) if any([f.endswith(".txt")])]
    num_file = 0
    for f in readmes:
        if f.__contains__("%s_" % readme_name):
            temp = f.split(".txt")[0]
            num_file = max(num_file, int(temp.split("%s_" % readme_name)[1]))
        else:
            num_file = 1
    readme_name = "ReadMe_%s.txt" % (num_file + 1)
    readme = open("%s%s/%s" % (path, out_dir, readme_name), "w")
    readme.write(update)
    readme.close()

    # ------------------------------------------------------------------------
    # setup all the slicers. set up random_seed for random/repRandom strategies through stacker_list.
    slicer = {}
    stacker_list = {}

    if specified_dith is not None:
        # would like to add all the stackers first and then keep only the one that is specified
        best_dith_only, no_dith_only = False, False

    if best_dith_only:
        stacker_list["RandomDitherFieldPerVisit"] = [
            mafStackers.RandomDitherFieldPerVisitStacker(degrees=ra_dec_in_deg, random_seed=1000)
        ]
        slicer["RandomDitherFieldPerVisit"] = slicers.HealpixSlicer(
            lon_col="randomDitherFieldPerVisitRa",
            lat_col="randomDitherFieldPerVisitDec",
            lat_lon_deg=ra_dec_in_deg,
            nside=nside,
            use_cache=False,
        )
    else:
        slicer["NoDither"] = slicers.HealpixSlicer(
            lon_col="fieldRA",
            lat_col="fieldDec",
            lat_lon_deg=ra_dec_in_deg,
            nside=nside,
            use_cache=False,
        )
        if not no_dith_only:
            # random dithers on different timescales
            stacker_list["RandomDitherPerNight"] = [
                mafStackers.RandomDitherPerNightStacker(degrees=ra_dec_in_deg, random_seed=1000)
            ]
            stacker_list["RandomDitherFieldPerNight"] = [
                mafStackers.RandomDitherFieldPerNightStacker(degrees=ra_dec_in_deg, random_seed=1000)
            ]
            stacker_list["RandomDitherFieldPerVisit"] = [
                mafStackers.RandomDitherFieldPerVisitStacker(degrees=ra_dec_in_deg, random_seed=1000)
            ]

            # rep random dithers on different timescales
            # stacker_list['RepulsiveRandomDitherPerNight'] = [myStackers.RepulsiveRandomDitherPerNightStacker(degrees=ra_dec_in_deg,
            #                                                                                                random_seed=1000)]
            # stacker_list['RepulsiveRandomDitherFieldPerNight'] = [myStackers.RepulsiveRandomDitherFieldPerNightStacker(degrees=ra_dec_in_deg,
            #                                                                                                          random_seed=1000)]
            # stacker_list['RepulsiveRandomDitherFieldPerVisit'] = [myStackers.RepulsiveRandomDitherFieldPerVisitStacker(degrees=ra_dec_in_deg,
            #                                                                                                          random_seed=1000)]
            # set up slicers for different dithers
            # random dithers on different timescales
            slicer["RandomDitherPerNight"] = slicers.HealpixSlicer(
                lon_col="randomDitherPerNightRa",
                lat_col="randomDitherPerNightDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            slicer["RandomDitherFieldPerNight"] = slicers.HealpixSlicer(
                lon_col="randomDitherFieldPerNightRa",
                lat_col="randomDitherFieldPerNightDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            slicer["RandomDitherFieldPerVisit"] = slicers.HealpixSlicer(
                lon_col="randomDitherFieldPerVisitRa",
                lat_col="randomDitherFieldPerVisitDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            # rep random dithers on different timescales
            # slicer['RepulsiveRandomDitherPerNight'] = slicers.HealpixSlicer(lonCol='repulsiveRandomDitherPerNightRa',
            #                                                               latCol='repulsiveRandomDitherPerNightDec',
            #                                                               latLonDeg=ra_dec_in_deg, nside=nside, use_cache=False)
            # slicer['RepulsiveRandomDitherFieldPerNight'] = slicers.HealpixSlicer(lonCol='repulsiveRandomDitherFieldPerNightRa',
            #                                                                    latCol='repulsiveRandomDitherFieldPerNightDec',
            #                                                                    latLonDeg=ra_dec_in_deg, nside=nside,
            #                                                                    use_cache=False)
            # slicer['RepulsiveRandomDitherFieldPerVisit'] = slicers.HealpixSlicer(lonCol='repulsiveRandomDitherFieldPerVisitRa',
            #                                                                    latCol='repulsiveRandomDitherFieldPerVisitDec',
            #                                                                    latLonDeg=ra_dec_in_deg, nside=nside,
            #                                                                    use_cache=False)
            # spiral dithers on different timescales
            slicer["FermatSpiralDitherPerNight"] = slicers.HealpixSlicer(
                lon_col="fermatSpiralDitherPerNightRa",
                lat_col="fermatSpiralDitherPerNightDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            slicer["FermatSpiralDitherFieldPerNight"] = slicers.HealpixSlicer(
                lon_col="fermatSpiralDitherFieldPerNightRa",
                lat_col="fermatSpiralDitherFieldPerNightDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            slicer["FermatSpiralDitherFieldPerVisit"] = slicers.HealpixSlicer(
                lon_col="fermatSpiralDitherFieldPerVisitRa",
                lat_col="fermatSpiralDitherFieldPerVisitDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            # hex dithers on different timescales
            slicer["SequentialHexDitherPerNight"] = slicers.HealpixSlicer(
                lon_col="hexDitherPerNightRa",
                lat_col="hexDitherPerNightDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            slicer["SequentialHexDitherFieldPerNight"] = slicers.HealpixSlicer(
                lon_col="hexDitherFieldPerNightRa",
                lat_col="hexDitherFieldPerNightDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            slicer["SequentialHexDitherFieldPerVisit"] = slicers.HealpixSlicer(
                lon_col="hexDitherFieldPerVisitRa",
                lat_col="hexDitherFieldPerVisitDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            # per season dithers
            slicer["PentagonDitherPerSeason"] = slicers.HealpixSlicer(
                lon_col="pentagonDitherPerSeasonRa",
                lat_col="pentagonDitherPerSeasonDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            slicer["PentagonDiamondDitherPerSeason"] = slicers.HealpixSlicer(
                lon_col="pentagonDiamondDitherPerSeasonRa",
                lat_col="pentagonDiamondDitherPerSeasonDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            slicer["SpiralDitherPerSeason"] = slicers.HealpixSlicer(
                lon_col="spiralDitherPerSeasonRa",
                lat_col="spiralDitherPerSeasonDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
    # ------------------------------------------------------------------------
    if specified_dith is not None:
        stacker_list_, slicer_ = {}, {}
        if isinstance(specified_dith, str):
            if specified_dith in slicer.keys():
                if specified_dith.__contains__("Random"):
                    # only Random dithers have a stacker object for rand seed specification
                    stacker_list_[specified_dith] = stacker_list[specified_dith]
                slicer_[specified_dith] = slicer[specified_dith]
        elif isinstance(specified_dith, list):
            for specific in specified_dith:
                if specific in slicer.keys():
                    if specific.__contains__("Random"):
                        # only Random dithers have a stacker object for rand seed specification
                        stacker_list_[specific] = stacker_list[specific]
                    slicer_[specific] = slicer[specific]
        else:
            err = "Invalid value for specified_dith: %s." % specified_dith
            err += "Allowed values include one of the following:\n%s" % (slicer.keys())
            raise ValueError(err)
        stacker_list, slicer = stacker_list_, slicer_

    print("\nRunning the analysis for %s" % slicer.keys())
    # ------------------------------------------------------------------------
    readme = open("%s%s/%s" % (path, out_dir, readme_name), "a")
    readme.write("\nObserving strategies considered: %s\n" % (list(slicer.keys())))
    readme.close()
    # ------------------------------------------------------------------------
    # set up bundle for num_gal (and later deltaN/N)
    my_bundles = {}
    dust_map = maps.DustMap(
        interp=False, nside=nside
    )  # include dust_map; actual in/exclusion of dust is handled by the galaxyCountMetric
    for dither in slicer:
        if dither in stacker_list:
            my_bundles[dither] = metricBundles.MetricBundle(
                gal_count_metric,
                slicer[dither],
                sqlconstraint,
                stacker_list=stacker_list[dither],
                run_name=run_name,
                metadata=dither,
                maps_list=[dust_map],
            )
        else:
            my_bundles[dither] = metricBundles.MetricBundle(
                gal_count_metric,
                slicer[dither],
                sqlconstraint,
                run_name=run_name,
                metadata=dither,
                maps_list=[dust_map],
            )
    # ------------------------------------------------------------------------
    # run the metric/slicer combination for galaxy counts (num_gal)
    print("\n# Running my_bundles ...")
    b_group = metricBundles.MetricBundleGroup(
        my_bundles,
        opsdb,
        out_dir="%s%s" % (path, out_dir),
        results_db=results_db,
        save_early=False,
    )
    b_group.run_all()
    b_group.write_all()

    # ------------------------------------------------------------------------
    # print out tot(num_gal) associated with each strategy
    # write to the readme as well
    update = "\n# Before any border masking or photometric error calibration: "
    print(update)
    for dither in my_bundles:
        ind = np.where(my_bundles[dither].metricValues.mask[:] == False)[0]
        print_out = "Total Galaxies for %s: %.9e" % (
            dither,
            sum(my_bundles[dither].metricValues.data[ind]),
        )
        update += "\n %s" % print_out
        print(print_out)
    update += "\n"
    readme = open("%s%s/%s" % (path, out_dir, readme_name), "a")
    readme.write(update)
    readme.close()
    print("\n## Time since the start of the calculation: %.2f hrs" % ((time.time() - start_time) / 3600.0))

    # ------------------------------------------------------------------------
    # mask the edges: the data in the masked pixels is not changed
    plot_handler = plots.PlotHandler(
        out_dir="%s%s" % (path, out_dir),
        results_db=results_db,
        thumbnail=False,
        savefig=False,
    )
    print("\n# Masking the edges ...")
    my_bundles, border_pixels_masked = masking_algorithm_generalized(
        my_bundles,
        plot_handler,
        "Number of Galaxies",
        nside=nside,
        pixel_radius=pixel_radius_for_masking,
        plot_intermediate_plots=False,
        plot_final_plots=False,
        print_final_info=True,
        return_border_indices=True,
    )
    # ------------------------------------------------------------------------

    # save the num_gal data.
    if save_num_gal_data_after_masking:
        out_dir_new = "numGalData_afterBorderMasking"
        for b in my_bundles:
            my_bundles[b].write(out_dir=out_dir_new)

    # ------------------------------------------------------------------------
    # print out tot(num_gal) associated with each strategy
    # write to the readme as well
    if pixel_radius_for_masking != 0:
        update = "\n# After border masking: "
        print(update)
        for dither in my_bundles:
            ind = np.where(my_bundles[dither].metricValues.mask[:] == False)[0]
            print_out = "Total Galaxies for %s: %.9e" % (
                dither,
                sum(my_bundles[dither].metricValues.data[ind]),
            )
            print(print_out)
            update += "\n %s" % print_out
        update += "\n"

        readme = open("%s%s/%s" % (path, out_dir, readme_name), "a")
        readme.write(update)
        readme.close()
    print("\n## Time since the start of the calculation: %.2f hrs" % ((time.time() - start_time) / 3600.0))

    ################################################################################################################
    # If include 0pt errors
    # Ansatz: for each pixel i, del_i= k*z_i/sqrt(n_obs_i),
    # where z_i is the average seeing the pixel minus avgSeeing across map, nObs is the number of observations,
    # and k is a constant such that var(del_i)= (0.01)^2. 0.01 for the 1% LSST goal.
    # k-constraint equation becomes: k^2*var(z_i/sqrt(n_obs_i))= (0.01)^2    --- equation 1
    if include0pt_errors:
        tablename = "SummaryAllProps"
        if tablename in opsdb.tableNames:
            colname = "seeingFwhmEff"
            if colname not in opsdb.columnNames[tablename]:
                raise ValueError("Unclear which seeing column to use.")
        elif "Summary" in opsdb.tableNames:
            tablename = "Summary"
            colname = "finSeeing"
            if colname not in opsdb.columnNames[tablename]:
                colname = "FWHMeff"
                if colname not in opsdb.columnNames[tablename]:
                    raise ValueError("Unclear which seeing column to use.")

        mean_metric = metrics.MeanMetric(col=colname)  # for avgSeeing per HEALpix pixel

        n_obs_metric = NumObsMetric(nside=nside)  # for numObs per HEALpix pixel
        if include_dust_extinction:
            coadd_metric = metrics.ExgalM5(lsstFilter=filter_band)
        else:
            coadd_metric = metrics.Coaddm5Metric()

        avg_seeing_bundle = {}
        n_obs_bundle = {}
        coadd_bundle = {}

        # can pass dust_map to metricBundle regardless of whether to include dust extinction or not.
        # the metric choice (coadd vs. exGal) takes care of whether to use the dust_map or not.
        dust_map = maps.DustMap(interp=False, nside=nside)
        for dither in slicer:
            if dither in stacker_list:
                avg_seeing_bundle[dither] = metricBundles.MetricBundle(
                    mean_metric,
                    slicer[dither],
                    sqlconstraint,
                    stacker_list=stacker_list[dither],
                    run_name=run_name,
                    metadata=dither,
                )
                n_obs_bundle[dither] = metricBundles.MetricBundle(
                    n_obs_metric,
                    slicer[dither],
                    sqlconstraint,
                    stacker_list=stacker_list[dither],
                    run_name=run_name,
                    metadata=dither,
                )
                coadd_bundle[dither] = metricBundles.MetricBundle(
                    coadd_metric,
                    slicer[dither],
                    sqlconstraint,
                    stacker_list=stacker_list[dither],
                    run_name=run_name,
                    metadata=dither,
                    maps_list=[dust_map],
                )
            else:
                avg_seeing_bundle[dither] = metricBundles.MetricBundle(
                    mean_metric,
                    slicer[dither],
                    sqlconstraint,
                    run_name=run_name,
                    metadata=dither,
                )
                n_obs_bundle[dither] = metricBundles.MetricBundle(
                    n_obs_metric,
                    slicer[dither],
                    sqlconstraint,
                    run_name=run_name,
                    metadata=dither,
                )
                coadd_bundle[dither] = metricBundles.MetricBundle(
                    coadd_metric,
                    slicer[dither],
                    sqlconstraint,
                    run_name=run_name,
                    metadata=dither,
                    maps_list=[dust_map],
                )
        print("\n# Running avg_seeing_bundle ...")
        a_group = metricBundles.MetricBundleGroup(
            avg_seeing_bundle,
            opsdb,
            out_dir="%s%s" % (path, out_dir),
            results_db=results_db,
            save_early=False,
        )
        a_group.run_all()

        print("\n# Running n_obs_bundle ...")
        n_group = metricBundles.MetricBundleGroup(
            n_obs_bundle,
            opsdb,
            out_dir="%s%s" % (path, out_dir),
            results_db=results_db,
            save_early=False,
        )
        n_group.run_all()

        print("\n# Running coadd_bundle ...")
        c_group = metricBundles.MetricBundleGroup(
            coadd_bundle,
            opsdb,
            out_dir="%s%s" % (path, out_dir),
            results_db=results_db,
            save_early=False,
        )
        c_group.run_all()

        # ------------------------------------------------------------------------
        # mask the border pixels
        for dither in slicer:
            avg_seeing_bundle[dither].metricValues.mask[border_pixels_masked[dither]] = True
            n_obs_bundle[dither].metricValues.mask[border_pixels_masked[dither]] = True
            coadd_bundle[dither].metricValues.mask[border_pixels_masked[dither]] = True

        # ------------------------------------------------------------------------
        # calculate averageSeeing over the entrie map
        bundle = {}
        bundle["avg_seeing_across_map"] = metricBundles.MetricBundle(
            mean_metric,
            slicers.UniSlicer(),
            sqlconstraint,
            run_name=run_name,
            metadata="avg_seeing_across_map",
        )
        bundle_group = metricBundles.MetricBundleGroup(
            bundle,
            opsdb,
            out_dir="%s%s" % (path, out_dir),
            results_db=results_db,
            save_early=False,
        )
        bundle_group.run_all()
        avg_seeing_across_map = bundle["avg_seeing_across_map"].metricValues.data[0]
        print_out = "\n# Average seeing across map: %s" % (avg_seeing_across_map)
        print(print_out)

        # add to the readme
        readme = open("%s%s/%s" % (path, out_dir, readme_name), "a")
        readme.write(print_out)
        readme.close()

        # find the zero point uncertainties: for each pixel i, del_i=k*z_i/sqrt(n_obs_i),
        # where z_i is the average seeing the pixel minus avgSeeing across map, nObs is the number of observations,
        # and k is a constant such that var(del_i)=(0.01)^2.
        # k-constraint equation becomes: k^2*var(z_i/sqrt(n_obs_i))=(0.01)^2    --- equation 1
        k = Symbol("k")
        zero_pt_error = {}
        k_value = {}

        print(
            "\n# 0pt calculation ansatz: \delta_i=k*z_i/sqrt{n_obs_i}, where k is s.t. var(\delta_i)=(0.01)^$"
        )

        if save0pt_plots:
            out_dir_new = "0pt_plots"
            if not os.path.exists("%s%s/%s" % (path, out_dir, out_dir_new)):
                os.makedirs("%s%s/%s" % (path, out_dir, out_dir_new))

        # ------------------------------------------------------------------------
        # add to the readme
        readme = open("%s%s/%s" % (path, out_dir, readme_name), "a")
        readme.write("\n\n0pt Information: ")
        readme.close()

        for dither in avg_seeing_bundle:
            z_i = avg_seeing_bundle[dither].metricValues.data[:] - avg_seeing_across_map
            n_obs_i = n_obs_bundle[dither].metricValues.data[:]
            ind = np.where((n_obs_bundle[dither].metricValues.mask == False) & (n_obs_i != 0.0))[
                0
            ]  # make sure the uncertainty is valid; no division by 0
            temp = np.var(z_i[ind] / np.sqrt(n_obs_i[ind]))  # see equation 1
            k_value[dither] = solve(k**2 * temp - 0.01**2, k)[1]

            err = np.empty(len(z_i))
            err.fill(-500)  # initiate
            err[ind] = (k_value[dither] * z_i[ind]) / np.sqrt(n_obs_i[ind])
            zero_pt_error[dither] = err

            # add to the readme
            readme = open("%s%s/%s" % (path, out_dir, readme_name), "a")
            readme.write("\nDith strategy: %s" % dither)
            readme.close()

            # ------------------------------------------------------------------------
            if print0pt_information:
                update = "\n# %s" % dither
                ind = np.where(zero_pt_error[dither] != -500)[0]
                good_error = zero_pt_error[dither][ind]
                update += "var(0pt): %s" % np.var(good_error)
                update += "\n0.01^2 - var(0pt) = %s" % ((0.01) ** 2 - np.var(good_error))
                update += "\nk-value: %s\n" % k_value[dither]
                print(update)
                # add to the readme
                readme = open("%s%s/%s" % (path, out_dir, readme_name), "a")
                readme.write(update)
                readme.close()
            # ------------------------------------------------------------------------
            if plot0pt_plots:
                # since not saving the bundle for 0pt errors, must plot out stuff without the plotBundle routine.
                ind = np.where(zero_pt_error[dither] != -500)[0]
                good_error = zero_pt_error[dither][ind]

                for i in range(len(good_error)):
                    good_error[i] = float(good_error[i])

                update = "\n# %s" % dither
                update += "\nMin error: %s" % min(good_error)
                update += "\nMax error: %s" % max(good_error)
                update += "\nMean error: %s" % np.mean(good_error)
                update += "\nStd of error: %s\n" % np.std(good_error)
                print(update)

                # add to the readme
                readme = open("%s%s/%s" % (path, out_dir, readme_name), "a")
                readme.write(update)
                readme.close()

                # plot histogram
                bin_size = 0.005
                bins = np.arange(
                    min(good_error) - 5 * bin_size,
                    max(good_error) + 5 * bin_size,
                    bin_size,
                )
                plt.clf()
                plt.hist(good_error, bins=bins)
                plt.xlabel("Zeropoint Uncertainty")
                plt.ylabel("Counts")

                plt.title(
                    "0pt error histogram; bin_size = %s; upper_mag_limit = %s" % (bin_size, upper_mag_limit)
                )
                if save0pt_plots:
                    filename = "0ptHistogram_%s_%s.png" % (filter_band, dither)
                    plt.savefig(
                        "%s%s/%s/%s" % (path, out_dir, out_dir_new, filename),
                        format="png",
                    )
                if show0pt_plots:
                    plt.show()
                else:
                    plt.close()

                # plot skymap
                temp = copy.deepcopy(coadd_bundle[dither])
                temp.metricValues.data[ind] = good_error
                temp.metricValues.mask[:] = True
                temp.metricValues.mask[ind] = False

                in_survey_index = np.where(temp.metricValues.mask == False)[0]
                median = np.median(temp.metricValues.data[in_survey_index])
                stddev = np.std(temp.metricValues.data[in_survey_index])

                color_min = -0.010  # median-1.5*stddev
                color_max = 0.010  # median+1.5*stddev
                n_ticks = 5
                increment = (color_max - color_min) / float(n_ticks)
                ticks = np.arange(color_min + increment, color_max, increment)

                plt.clf()
                hp.mollview(
                    temp.metricValues.filled(temp.slicer.badval),
                    flip="astro",
                    rot=(0, 0, 0),
                    min=color_min,
                    max=color_max,
                    title="",
                    cbar=False,
                )
                hp.graticule(dpar=20, dmer=20, verbose=False)
                plt.title(dither)
                ax = plt.gca()
                im = ax.get_images()[0]
                fig = plt.gcf()
                cbaxes = fig.add_axes([0.1, 0.03, 0.8, 0.04])  # [left, bottom, width, height]
                cb = plt.colorbar(im, orientation="horizontal", ticks=ticks, format="%.3f", cax=cbaxes)
                cb.set_label("Photometric Calibration Error")

                if save0pt_plots:
                    filename = "0ptSkymap_%s.png" % (dither)
                    plt.savefig(
                        "%s%s/%s/%s" % (path, out_dir, out_dir_new, filename),
                        bbox_inches="tight",
                        format="png",
                    )

                if show0pt_plots:
                    plt.show()
                else:
                    plt.close()

                # plot power spectrum
                plt.clf()
                spec = hp.anafast(temp.metricValues.filled(temp.slicer.badval), lmax=500)
                ell = np.arange(np.size(spec))
                condition = ell > 1
                plt.plot(ell, (spec * ell * (ell + 1)) / 2.0 / np.pi)
                plt.title("Photometric Calibration Error: %s" % dither)
                plt.xlabel(r"$\ell$")
                plt.ylabel(r"$\ell(\ell+1)C_\ell/(2\pi)$")
                plt.xlim(0, 500)

                if save0pt_plots:
                    # save power spectrum
                    filename = "0ptPowerSpectrum_%s.png" % (dither)
                    plt.savefig(
                        "%s%s/%s/%s" % (path, out_dir, out_dir_new, filename),
                        bbox_inches="tight",
                        format="png",
                    )

                if show0pt_plots:
                    plt.show()
                else:
                    plt.close()

        print(
            "\n## Time since the start of the calculation: %.2f hrs" % ((time.time() - start_time) / 3600.0)
        )

        # ------------------------------------------------------------------------
        # Now recalculate the num_gal with the fluctuations in depth due to calibation uncertainties.
        print("\n# Recalculating num_gal including 0pt errors on the upper mag limit .. ")
        for dither in my_bundles:
            zero_pt_err = zero_pt_error[dither].copy()
            in_survey = np.where(my_bundles[dither].metricValues.mask == False)[
                0
            ]  # 04/27: only look at in_survey region
            for i in in_survey:  # 4/27
                if zero_pt_err[i] != -500:  # run only when zeroPt was calculated
                    my_bundles[dither].metricValues.data[i] = GalaxyCounts_0ptErrors(
                        coadd_bundle[dither].metricValues.data[i],
                        upper_mag_limit + zero_pt_err[i],
                        redshift_bin=redshift_bin,
                        filter_band=filter_band,
                        nside=nside,
                        cfhtls_counts=cfhtls_counts,
                        normalized_mock_catalog_counts=normalized_mock_catalog_counts,
                    )
        # ------------------------------------------------------------------------

        # save the raw num_gal data.
        if save_num_gal_data_after0pt:
            out_dir_new = "numGalData_afterBorderMasking_after0pt"
            for b in my_bundles:
                my_bundles[b].write(out_dir=out_dir_new)

        # ------------------------------------------------------------------------
        # print out tot(num_gal) associated with each strategy
        # add to the read me as well
        update = "\n# After 0pt error calculation and border masking: "
        print(update)
        for dither in my_bundles:
            ind = np.where(my_bundles[dither].metricValues.mask[:] == False)[0]
            print_out = "Total Galaxies for %s: %.9e" % (
                dither,
                sum(my_bundles[dither].metricValues.data[ind]),
            )
            update += "\n %s" % print_out
            print(print_out)
        update += "\n"
        readme = open("%s%s/%s" % (path, out_dir, readme_name), "a")
        readme.write(update)
        readme.close()

    print("\n## Time since the start of the calculation: %.2f hrs" % ((time.time() - start_time) / 3600.0))

    #########################################################################################################
    # add poisson noise?
    if add_poisson_noise:
        print("\n# adding poisson noise to num_gal ... ")
        for dither in my_bundles:
            # make sure the values are valid; sometimes metric leaves negative numbers or nan values.
            out_of_survey = np.where(my_bundles[dither].metricValues.mask == True)[0]
            my_bundles[dither].metricValues.data[out_of_survey] = 0.0

            in_survey = np.where(my_bundles[dither].metricValues.mask == False)[0]
            j = np.where(my_bundles[dither].metricValues.data[in_survey] < 1.0)[0]
            my_bundles[dither].metricValues.data[in_survey][j] = 0.0

            noisy_num_gal = np.random.poisson(lam=my_bundles[dither].metricValues.data, size=None)
            my_bundles[dither].metricValues.data[:] = noisy_num_gal
        # ------------------------------------------------------------------------

        # save the num_gal data.
        if saveNumGalDataAfterPoisson:
            out_dir_new = "numGalData_afterBorderMasking_after0pt_afterPoisson"
            for b in my_bundles:
                my_bundles[b].write(out_dir=out_dir_new)

        # ------------------------------------------------------------------------
        # print out tot(num_gal) associated with each strategy
        # add to the read me as well
        update = "\n# After adding poisson noise: "
        print(update)
        for dither in my_bundles:
            ind = np.where(my_bundles[dither].metricValues.mask[:] == False)[0]
            print_out = "Total Galaxies for %s: %.9e" % (
                dither,
                sum(my_bundles[dither].metricValues.data[ind]),
            )
            update += "\n %s" % print_out
            print(print_out)
        update += "\n"
        readme = open("%s%s/%s" % (path, out_dir, readme_name), "a")
        readme.write(update)
        readme.close()

    print("\n## Time since the start of the calculation: %.2f hrs" % ((time.time() - start_time) / 3600.0))
    #########################################################################################################
    plot_handler = plots.PlotHandler(
        out_dir="%s%s" % (path, out_dir),
        results_db=results_db,
        thumbnail=False,
        savefig=False,
    )
    print("\n# Calculating fluctuations in the galaxy counts ...")
    # Change num_gal metric data to deltaN/N
    num_gal = {}
    # add to readme too
    update = "\n"
    for dither in my_bundles:
        # zero out small/nan entries --- problem: should really be zeroed out by the metric ***
        j = np.where(np.isnan(my_bundles[dither].metricValues.data) == True)[0]
        my_bundles[dither].metricValues.data[j] = 0.0
        j = np.where(my_bundles[dither].metricValues.data < 1.0)[0]
        my_bundles[dither].metricValues.data[j] = 0.0
        # calculate the fluctuations
        num_gal[dither] = my_bundles[
            dither
        ].metricValues.data.copy()  # keep track of num_gal for plotting purposes
        valid_pixel = np.where(my_bundles[dither].metricValues.mask == False)[0]
        galaxy_average = sum(num_gal[dither][valid_pixel]) / len(valid_pixel)

        # in place calculation of the fluctuations
        my_bundles[dither].metricValues.data[:] = 0.0
        my_bundles[dither].metricValues.data[valid_pixel] = (
            num_gal[dither][valid_pixel] - galaxy_average
        ) / galaxy_average
        print_out = "# Galaxy Average for %s: %s" % (dither, galaxy_average)
        print(print_out)
        update += "%s\n" % print_out

    readme = open("%s%s/%s" % (path, out_dir, readme_name), "a")
    readme.write(update)
    readme.close()

    # ------------------------------------------------------------------------
    # save the deltaN/N data
    if save_delta_n_by_n_data:
        out_dir_new = "deltaNByNData"
        for b in my_bundles:
            my_bundles[b].write(out_dir=out_dir_new)

    # ------------------------------------------------------------------------
    # Calculate total power
    # add to the read me as well
    summarymetric = metrics.TotalPowerMetric()
    update = ""
    for dither in my_bundles:
        my_bundles[dither].set_summary_metrics(summarymetric)
        my_bundles[dither].compute_summary_stats()
        print_out = "# Total power for %s case is %f." % (
            dither,
            my_bundles[dither].summary_values["TotalPower"],
        )
        print(print_out)
        update += "\n%s" % (print_out)
    update += "\n"

    readme = open("%s%s/%s" % (path, out_dir, readme_name), "a")
    readme.write(update)
    readme.close()
    # ------------------------------------------------------------------------
    # calculate the power spectra
    cl = {}
    for dither in my_bundles:
        cl[dither] = hp.anafast(
            my_bundles[dither].metricValues.filled(my_bundles[dither].slicer.badval),
            lmax=500,
        )
    # save deltaN/N spectra?
    if save_cls_for_delta_n_by_n:
        out_dir_new = "cls_DeltaByN"
        if not os.path.exists("%s%s/%s" % (path, out_dir, out_dir_new)):
            os.makedirs("%s%s/%s" % (path, out_dir, out_dir_new))

        for dither in my_bundles:
            filename = "cls_deltaNByN_%s_%s" % (filter_band, dither)
            np.save("%s%s/%s/%s" % (path, out_dir, out_dir_new, filename), cl[dither])

    ##########################################################################################################
    # Plots for the fluctuations: power spectra, histogram
    if len(list(my_bundles.keys())) > 1:
        out_dir_new = "artificialFluctuationsComparisonPlots"
        if not os.path.exists("%s%s/%s" % (path, out_dir, out_dir_new)):
            os.makedirs("%s%s/%s" % (path, out_dir, out_dir_new))
        # ------------------------------------------------------------------------
        # power spectra
        for dither in my_bundles:
            ell = np.arange(np.size(cl[dither]))
            condition = ell > 1
            plt.plot(
                ell,
                (cl[dither] * ell * (ell + 1)) / 2.0 / np.pi,
                color=plot_color[dither],
                linestyle="-",
                label=dither,
            )
        plt.xlabel(r"$\ell$")
        plt.ylabel(r"$\ell(\ell+1)C_\ell/(2\pi)$")
        plt.xlim(0, 500)
        leg = plt.legend(labelspacing=0.001)
        for legobj in leg.legendHandles:
            legobj.set_linewidth(2.0)
        plt.savefig(
            "%s%s/%s/powerspectrum_comparison.png" % (path, out_dir, out_dir_new),
            format="png",
        )
        if show_comp_plots:
            plt.show()
        else:
            plt.close("all")
        # ------------------------------------------------------------------------
        # create the histogram
        scale = hp.nside2pixarea(nside, degrees=True)

        def tick_formatter(y, pos):
            return "%d" % (y * scale)  # convert pixel count to area

        for dither in my_bundles:
            ind = np.where(my_bundles[dither].metricValues.mask == False)[0]
            bin_size = 0.01
            bin_all = int(
                (
                    max(my_bundles[dither].metricValues.data[ind])
                    - min(my_bundles[dither].metricValues.data[ind])
                )
                / bin_size
            )
            plt.hist(
                my_bundles[dither].metricValues.data[ind],
                bins=bin_all,
                label=dither,
                histtype="step",
                color=plot_color[dither],
            )
        # plt.xlim(-0.6,1.2)
        ax = plt.gca()
        ymin, ymax = ax.get_ylim()
        n_yticks = 10.0
        wanted_y_max = ymax * scale
        wanted_y_max = 10.0 * np.ceil(float(wanted_y_max) / 10.0)
        increment = 5.0 * np.ceil(float(wanted_y_max / n_yticks) / 5.0)
        wanted_array = np.arange(0, wanted_y_max, increment)
        ax.yaxis.set_ticks(wanted_array / scale)
        ax.yaxis.set_major_formatter(FuncFormatter(tick_formatter))
        plt.xlabel(r"$\mathrm{\Delta N/\overline{N}}$")
        plt.ylabel("Area (deg$^2$)")
        leg = plt.legend(labelspacing=0.001, bbox_to_anchor=(1, 1))
        for legobj in leg.legendHandles:
            legobj.set_linewidth(2.0)
        plt.savefig(
            "%s%s/%s/histogram_comparison.png" % (path, out_dir, out_dir_new),
            bbox_inches="tight",
            format="png",
        )
        if show_comp_plots:
            plt.show()
        else:
            plt.close("all")

    # now remove the results db object -- useless
    os.remove("%s%s/%s" % (path, out_dir, results_dbname))
    print("Removed %s from out_dir" % (results_dbname))

    # all done. final update.
    update = "\n## All done. Time since the start of the calculation: %.2f hrs" % (
        (time.time() - start_time) / 3600.0
    )
    print(update)
    readme = open("%s%s/%s" % (path, out_dir, readme_name), "a")
    readme.write(update)
    readme.close()

    if return_stuff:
        if include0pt_errors:
            return my_bundles, out_dir, results_db, zero_pt_error
        else:
            return my_bundles, out_dir, results_db



#!/usr/bin/env python

# Mike Lund - Vanderbilt University
# mike.lund@gmail.com
# Last edited 8/15/2015
# Description: Calculates the number of stars in a given direction and between a given set of distances. For use with Field Star Count metric
import math
import sys

import numpy as np
from scipy.optimize import fsolve

from . import coords, stellardensity

# from rubin_sim.coordUtils import AstronomyBase
skyarea = 41253.0
distancebins = 51


def star_vols(d1, d2, area):
    distance_edges = (np.linspace((d1**3.0), (d2**3.0), num=distancebins)) ** (1.0 / 3)
    volumeshell = (area / skyarea) * (4.0 * np.pi / 3) * (distance_edges[1:] ** 3 - distance_edges[:-1] ** 3)
    distances = ((distance_edges[1:] ** 3 + distance_edges[:-1] ** 3) / 2.0) ** (1.0 / 3)
    return volumeshell, distances


def starcount(eq_ra, eq_dec, d1, d2):
    volumes, distances = star_vols(d1, d2, 9.62)
    # b_deg, l_deg=coords.eq_gal2(eq_ra, eq_dec)
    # b_deg, l_deg=AstrometryBase.equatorialToGalactic(eq_ra, eq_dec)
    b_deg, l_deg = coords.eq_gal3(eq_ra, eq_dec)
    positions = [coords.gal_cyn(b_deg, l_deg, x) for x in distances]
    densities = [stellardensity.stellardensity(x[0], x[2]) for x in positions]
    totalcount = np.sum(np.asarray(volumes) * np.asarray(densities))
    return totalcount


if __name__ == "__main__":
    print(
        starcount(
            float(sys.argv[1]),
            float(sys.argv[2]),
            float(sys.argv[3]),
            float(sys.argv[4]),
        )
    )



__all__ = ("PlotBundle",)


import matplotlib.pylab as plt

from .plot_handler import PlotHandler


class PlotBundle:
    """
    Object designed to help organize multiple MetricBundles that will be plotted
    together using the PlotHandler.
    """

    def __init__(self, bundle_list=None, plot_dicts=None, plot_func=None):
        """
        Init object and set things if desired.
        bundle_list: A list of bundleDict objects
        plot_dicts: A list of dictionaries with plotting kwargs
        plot_func: A single MAF plotting function
        """
        if bundle_list is None:
            self.bundle_list = []
        else:
            self.bundle_list = bundle_list

        if plot_dicts is None:
            if len(self.bundle_list) > 0:
                self.plot_dicts = [{}]
            else:
                self.plot_dicts = []
        else:
            self.plot_dicts = plot_dicts

        self.plot_func = plot_func

    def add_bundle(self, bundle, plot_dict=None, plot_func=None):
        """
        Add bundle to the object.
        Optionally add a plot_dict and/or replace the plot_func
        """
        self.bundle_list.append(bundle)
        if plot_dict is not None:
            self.plot_dicts.append(plot_dict)
        else:
            self.plot_dicts.append({})
        if plot_func is not None:
            self.plot_func = plot_func

    def increment_plot_order(self):
        """
        Find the maximium order number in the display dicts, and set them to +1 that
        """
        max_order = 0
        for m_b in self.bundle_list:
            if "order" in list(m_b.displayDict.keys()):
                max_order = max([max_order, m_b.displayDict["order"]])

        for m_b in self.bundle_list:
            m_b.displayDict["order"] = max_order + 1

    def percentile_legend(self):
        """
        Go through the bundles and change the lables if there are the correct summary stats
        """
        for i, mB in enumerate(self.bundle_list):
            if mB.summary_values is not None:
                keys = list(mB.summary_values.keys())
                if ("25th%ile" in keys) & ("75th%ile" in keys) & ("Median" in keys):
                    if "label" not in list(self.plot_dicts[i].keys()):
                        self.plot_dicts[i]["label"] = ""
                    newstr = "%0.1f/%0.1f/%0.1f " % (
                        mB.summary_values["25th%ile"],
                        mB.summary_values["Median"],
                        mB.summary_values["75th%ile"],
                    )
                    self.plot_dicts[i]["label"] = newstr + self.plot_dicts[i]["label"]

    def plot(self, out_dir="Out", results_db=None, closefigs=True):
        ph = PlotHandler(out_dir=out_dir, results_db=results_db)
        ph.set_metric_bundles(self.bundle_list)
        # Auto-generate labels and things
        ph.set_plot_dicts(plot_dicts=self.plot_dicts, plot_func=self.plot_func)
        ph.plot(self.plot_func, plot_dicts=self.plot_dicts)
        if closefigs:
            plt.close("all")



__all__ = (
    "RunSelectHandler",
    "MetricSelectHandler",
    "MetricResultsPageHandler",
    "DataHandler",
    "ConfigPageHandler",
    "StatPageHandler",
    "AllMetricResultsPageHandler",
    "MultiColorPageHandler",
    "make_app",
    "show_maf",
)

import argparse
import os
import webbrowser

from jinja2 import Environment, FileSystemLoader
from tornado import ioloop, web

import rubin_sim
from rubin_sim.maf.db import add_run_to_database
from rubin_sim.maf.web import MafTracking


class RunSelectHandler(web.RequestHandler):
    def get(self):
        selectTempl = env.get_template("runselect.html")
        if "runId" in self.request.arguments:
            runId = int(self.request.arguments["runId"][0])
        else:
            # Set runID to a negative number, to default to first run.
            runId = startRunId
        self.write(selectTempl.render(runlist=runlist, runId=runId, jsPath=jsPath))


class MetricSelectHandler(web.RequestHandler):
    def get(self):
        selectTempl = env.get_template("metricselect.html")
        runId = int(self.request.arguments["runId"][0])
        self.write(selectTempl.render(runlist=runlist, runId=runId))


class MetricResultsPageHandler(web.RequestHandler):
    def get(self):
        resultsTempl = env.get_template("results.html")
        runId = int(self.request.arguments["runId"][0])
        if "metric_id" in self.request.arguments:
            metricIdList = self.request.arguments["metric_id"]
        else:
            metricIdList = []
        if "Group_subgroup" in self.request.arguments:
            groupList = self.request.arguments["Group_subgroup"]
        else:
            groupList = []
        self.write(
            resultsTempl.render(
                metricIdList=metricIdList,
                groupList=groupList,
                runId=runId,
                runlist=runlist,
            )
        )


class DataHandler(web.RequestHandler):
    def get(self):
        runId = int(self.request.arguments["runId"][0])
        metricId = int(self.request.arguments["metric_id"][0])
        if "datatype" in self.request.arguments:
            datatype = self.request.arguments["datatype"][0].lower()
            datatype = datatype.decode("utf-8")
        else:
            datatype = "npz"
        run = runlist.getRun(runId)
        metric = run.metricIdsToMetrics([metricId])
        if datatype == "npz":
            npz = run.getNpz(metric)
            if npz is None:
                self.write("No npz file available.")
            else:
                self.redirect(npz)
        elif datatype == "json":
            jsn = run.getJson(metric)
            if jsn is None:
                self.write("No JSON file available.")
            else:
                self.write(jsn)
        else:
            self.write('Data type "%s" not understood.' % (datatype))


class ConfigPageHandler(web.RequestHandler):
    def get(self):
        configTempl = env.get_template("configs.html")
        runId = int(self.request.arguments["runId"][0])
        self.write(configTempl.render(runlist=runlist, runId=runId))


class StatPageHandler(web.RequestHandler):
    def get(self):
        statTempl = env.get_template("stats.html")
        runId = int(self.request.arguments["runId"][0])
        self.write(statTempl.render(runlist=runlist, runId=runId))


class AllMetricResultsPageHandler(web.RequestHandler):
    def get(self):
        """Load up the files and display"""
        allresultsTempl = env.get_template("allmetricresults.html")
        runId = int(self.request.arguments["runId"][0])
        self.write(allresultsTempl.render(runlist=runlist, runId=runId))


class MultiColorPageHandler(web.RequestHandler):
    def get(self):
        """Display sky maps."""
        multiColorTempl = env.get_template("multicolor.html")
        runId = int(self.request.arguments["runId"][0])
        self.write(multiColorTempl.render(runlist=runlist, runId=runId))


def make_app():
    """The tornado global configuration"""
    application = web.Application(
        [
            ("/", RunSelectHandler),
            ("/metricSelect", MetricSelectHandler),
            ("/metricResults", MetricResultsPageHandler),
            ("/getData", DataHandler),
            ("/configParams", ConfigPageHandler),
            ("/summaryStats", StatPageHandler),
            ("/allMetricResults", AllMetricResultsPageHandler),
            ("/multiColor", MultiColorPageHandler),
            (r"/(favicon.ico)", web.StaticFileHandler, {"path": faviconPath}),
            (r"/(sorttable.js)", web.StaticFileHandler, {"path": jsPath}),
            (r"/*/(.*)", web.StaticFileHandler, {"path": staticpath}),
        ]
    )
    return application


def show_maf():
    """Display MAF output in a web browser. After launching, point your browser
    to 'http://localhost:8888/'.
    """

    parser = argparse.ArgumentParser(
        description="Python script to display MAF output in a web browser."
        + "  After launching, point your browser to 'http://localhost:8888/'"
    )
    defaultdb = os.path.join(os.getcwd(), "trackingDb_sqlite.db")
    parser.add_argument(
        "-t",
        "--tracking_db",
        type=str,
        default=defaultdb,
        help="Tracking database filename.",
    )
    parser.add_argument(
        "-d",
        "--maf_dir",
        type=str,
        default=None,
        help="Add this directory to the tracking_db and open immediately.",
    )
    parser.add_argument(
        "-c",
        "--maf_comment",
        type=str,
        default=None,
        help="Add a comment to the trackingDB describing the "
        + " MAF analysis of this directory (paired with maf_dir argument).",
    )
    parser.add_argument("-p", "--port", type=int, default=8888, help="Port for connecting to showMaf.")
    parser.add_argument(
        "--no_browser",
        dest="no_browser",
        default=False,
        action="store_true",
        help="Do not open a new browser tab",
    )

    args = parser.parse_args()

    # Check tracking DB is sqlite (and add as convenience if forgotten).
    tracking_db = args.tracking_db
    print("Using tracking database at %s" % (tracking_db))

    global startRunId
    startRunId = -666
    # If given a directory argument:
    if args.maf_dir is not None:
        maf_dir = os.path.realpath(args.maf_dir)
        if not os.path.isdir(maf_dir):
            print("There is no directory containing MAF outputs at %s." % (maf_dir))
            print("Just opening using tracking db at %s." % (tracking_db))
        else:
            add_run_to_database(maf_dir, tracking_db, maf_comment=args.maf_comment)

    # Open tracking database and start visualization.
    global runlist
    runlist = MafTracking(tracking_db)
    if startRunId < 0:
        startRunId = runlist.runs[0]["maf_run_id"]
    # Set up path to template and favicon paths, and load templates.
    maf_dir = os.path.join(rubin_sim.__path__[0], "maf")
    templateDir = os.path.join(maf_dir, "web/templates/")
    global faviconPath
    faviconPath = os.path.join(maf_dir, "web/")
    global jsPath
    jsPath = os.path.join(maf_dir, "web/")
    global env
    env = Environment(loader=FileSystemLoader(templateDir))
    # Add 'zip' to jinja templates.
    env.globals.update(zip=zip)

    global staticpath
    staticpath = "."

    # Start up tornado app.
    application = make_app()
    application.listen(args.port)
    print("Tornado Starting: \nPoint your web browser to http://localhost:%d/ \nCtrl-C to stop" % (args.port))
    if not args.no_browser:
        webbrowser.open_new_tab("http://localhost:%d" % (args.port))
    ioloop.IOLoop.instance().start()



#!/usr/bin/env python

# Mike Lund - Vanderbilt University
# mike.lund@gmail.com
# Last edited 8/15/2015
# Description: Calculates spectral types for stars on the main sequence as a function of stellar mass. For use with Field Star Count metric
import sys

import numpy as np
from scipy.interpolate import interp1d


def extrap1d(interpolator):
    xs = interpolator.x
    ys = interpolator.y

    def pointwise(x):
        if x < xs[0]:
            return ys[0]
        elif x > xs[-1]:
            return ys[-1]
        else:
            return interpolator(x)

    def ufunclike(xs):
        return np.array(list(map(pointwise, np.array(xs))))

    return ufunclike


def spec_type(mass):
    mass_range = [
        0.06,
        0.21,
        0.40,
        0.51,
        0.67,
        0.79,
        0.92,
        1.05,
        1.4,
        1.6,
        2.0,
        2.9,
        3.8,
        5.9,
        7.6,
        17.5,
        23,
        37,
        60,
        120,
    ]
    spec_range = [
        68,
        65,
        62,
        60,
        55,
        50,
        45,
        40,
        35,
        30,
        25,
        20,
        18,
        15,
        13,
        10,
        8,
        6,
        5,
        3,
    ]
    f = interp1d(mass_range, spec_range)
    f2 = extrap1d(f)
    # f2=interp1d(mass_range, spec_range, kind='slinear')
    if mass == 0:
        xnew = np.logspace(-2, 3, 100)
        import matplotlib.pyplot as plt

        plt.plot(mass_range, spec_range, "o", xnew, f2(xnew), "-")  # , xnew, f2(xnew),'--')
        plt.xscale("log")
        plt.show()
    return f2([mass])


if __name__ == "__main__":
    print(spec_type(float(sys.argv[1])))  # mass



#!/usr/bin/env python

__all__ = ("setup_args",)

import argparse
import logging
import os

import numpy as np

# So things don't fail on hyak
from astropy.utils import iers

import rubin_sim.moving_objects as mo
from rubin_sim.maf.batches import col_map_dict

iers.conf.auto_download = False


def setup_args(parser=None):
    """Parse the command line arguments.

    Parameters
    ----------
    parser: argparse.ArgumentParser, optional
        Generally left at the default (None), but a user could set up their own parser if desired.

    Returns
    -------
    argparse.Namespace
        The argument options.
    """

    if parser is None:
        parser = argparse.ArgumentParser(description="Generate moving object detections.")
    parser.add_argument(
        "--simulation_db",
        type=str,
        default=None,
        help="Simulation output db file (example: baseline_v2.1_10yrs.db). Default None.",
    )
    parser.add_argument(
        "--orbit_file",
        type=str,
        default=None,
        help="File containing the moving object orbits. "
        "See https://github.com/oorb/oorb/tree/master/python#defining-orbits for "
        "additional documentation on the orbit file format. Default None.",
    )
    parser.add_argument(
        "--positions_file",
        type=str,
        default=None,
        help="File with pre-computed 'rough' ephemerides for the objects in the orbit file."
        " Default None, which will trigger computation of these rough ephemerides on-the-fly.",
    )
    parser.add_argument(
        "--out_dir",
        type=str,
        default=".",
        help="Output directory for moving object detections. Default '.'",
    )
    parser.add_argument(
        "--output_file",
        type=str,
        default=None,
        help="Output file name for moving object observations."
        " Default will build out_dir/simulation_db_orbitFile_obs.npz.",
    )
    parser.add_argument(
        "--sql_constraint",
        type=str,
        default="",
        help="SQL constraint to use to select data from simulation_db. Default no constraint.",
    )
    parser.add_argument(
        "--obs_info",
        type=str,
        default=None,
        help="Additional info to write into output file. "
        "The default info will combine the simulation_db name, the sqlconstraint, and "
        "the name of the orbit file; obs_info is an optional addition.",
    )
    parser.add_argument(
        "--footprint",
        type=str,
        default="camera",
        help="Type of footprint to use to identify observations of each object. "
        "Options are 'circle' (r=1.75 deg), 'rectangle', or 'camera' (camera footprint). "
        "Default is 'camera'.",
    )
    parser.add_argument(
        "--r_fov",
        type=float,
        default=1.75,
        help="If using a circular footprint, this is the radius of the FOV (in degrees). "
        "Default 1.75 deg.",
    )
    parser.add_argument(
        "--x_tol",
        type=float,
        default=5,
        help="If using a rectangular footprint, this is the tolerance in the RA direction "
        "(in degrees). Default is 5 degrees.",
    )
    parser.add_argument(
        "--y_tol",
        type=float,
        default=3,
        help="If using a rectangular footprint, this is the tolerance in the Dec direction "
        "(in degrees). Default is 3 degrees.",
    )
    parser.add_argument(
        "--rough_tol",
        type=float,
        default=10,
        help="If using direct/exact ephemeris generation, this is the tolerance for the "
        "preliminary matches between ephemerides and pointings (in degrees). "
        "Default 10 degrees.",
    )
    parser.add_argument(
        "--obs_code",
        type=str,
        default="I11",
        help="Observatory code for generating observations. " "Default is I11 (Cerro Pachon).",
    )
    parser.add_argument(
        "--t_step",
        type=float,
        default=1.0,
        help="Timestep between ephemeris generation for either the first (rough) stage of "
        "direct ephemeris generation or the grid for linear interpolation "
        "ephemerides. Default 1 day.",
    )
    parser.add_argument(
        "--eph_mode",
        type=str,
        default="nbody",
        help="2body or nbody mode for ephemeris generation. Default is nbody.",
    )
    parser.add_argument(
        "--prelim_eph_mode",
        type=str,
        default="nbody",
        help="Use either 2body or nbody for preliminary ephemeris generation in the rough "
        "stage for DirectObs. Default 2body.",
    )
    parser.add_argument(
        "--eph_type",
        type=str,
        default="basic",
        help="Generate either 'basic' or 'full' ephemerides from OOrb. "
        "See https://github.com/lsst/oorb/blob/lsst-dev/python/README.rst for details"
        "of the contents of 'full' or 'basic' ephemerides. "
        "Default basic.",
    )
    parser.add_argument(
        "--log_file",
        type=str,
        default=None,
        help="Send log output to log_file, instead of to console. (default = console)",
    )
    args = parser.parse_args()

    if args.simulation_db is None:
        raise ValueError("Must specify an simulation database output file.")

    if args.orbit_file is None:
        raise ValueError("Must specify an orbit file.")

    run_name = os.path.split(args.simulation_db)[-1].replace(".db", "")

    # Add these useful pieces to args.
    args.orbitbase = ".".join(os.path.split(args.orbit_file)[-1].split(".")[:-1])

    # Set up obs_file if not specified.
    if args.output_file is None:
        args.output_file = os.path.join(args.out_dir, "%s__%s_obs.npz" % (run_name, args.orbitbase))
    else:
        args.output_file = os.path.join(args.out_dir, args.output_file)

    # Build some provenance to add to output file.
    obs_info = args.simulation_db
    if len(args.sql_constraint) > 0:
        obs_info += " selected with sqlconstraint %s" % args.sql_constraint
    obs_info += " + orbit_file %s" % args.orbitbase
    if args.obs_info is not None:
        obs_info += "\n# %s" % args.obs_info
    args.obs_info = obs_info

    return args


def make_lsst_obs():
    # Parser command
    args = setup_args()

    # Send info and above logging messages to the console or logfile.
    if args.log_file is not None:
        logging.basicConfig(filename=args.log_file, level=logging.INFO)
    else:
        logging.basicConfig(level=logging.INFO)

    # Read orbits.
    orbits = mo.Orbits()
    orbits.read_orbits(args.orbit_file)

    if args.positions_file is not None:
        position_data = np.load(args.positions_file)
        object_positions = position_data["positions"].copy()
        object_mjds = position_data["mjds"].copy()
        position_data.close()
    else:
        object_positions = None
        object_mjds = None
    # Read pointing data
    colmap = col_map_dict("fbs")
    pointing_data = mo.read_observations(
        args.simulation_db,
        colmap,
        constraint=args.sql_constraint,
        dbcols=None,
    )

    d_obs = mo.DirectObs(
        footprint=args.footprint,
        r_fov=args.r_fov,
        x_tol=args.x_tol,
        y_tol=args.y_tol,
        eph_mode=args.eph_mode,
        prelim_eph_mode=args.prelim_eph_mode,
        obs_code=args.obs_code,
        eph_file=None,
        eph_type=args.eph_type,
        obs_time_col=colmap["mjd"],
        obs_time_scale="TAI",
        seeing_col=colmap["seeingGeom"],
        visit_exp_time_col=colmap["exptime"],
        obs_ra=colmap["ra"],
        obs_dec=colmap["dec"],
        obs_rot_sky_pos=colmap["rotSkyPos"],
        obs_degrees=colmap["raDecDeg"],
        outfile_name=None,
        tstep=args.t_step,
        rough_tol=args.rough_tol,
        obs_info=args.obs_info,
    )
    filterlist = np.unique(pointing_data["filter"])
    d_obs.read_filters(filterlist=filterlist)
    # Calculate all colors ahead of time.
    sednames = np.unique(orbits.orbits["sed_filename"])
    for sedname in sednames:
        d_obs.calc_colors(sedname)

    # Generate object observations.
    object_observations = d_obs.run(
        orbits,
        pointing_data,
        object_positions=object_positions,
        object_mjds=object_mjds,
    )

    np.savez(
        os.path.join(args.out_dir, args.output_file),
        object_observations=object_observations,
        info=d_obs.info,
    )

    # logging.info("Completed successfully.")



#####################################################################################################
# Purpose: change the values/mask of a metricBundle in the pixels with a certain value/mask.
# Example applicaton: mask the outermost/shallow edge of skymaps.
#
# Humna Awan: humna.awan@rutgers.edu
#####################################################################################################

__all__ = ("masking_algorithm_generalized",)

import copy

import healpy as hp
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np

import rubin_sim.maf.plots as plots


def masking_algorithm_generalized(
    my_bundles,
    plot_handler,
    data_label,
    nside=128,
    find_value="unmasked",
    relation="=",
    new_value="masked",
    pixel_radius=6,
    return_border_indices=False,
    print_intermediate_info=False,
    plot_intermediate_plots=True,
    print_final_info=True,
    plot_final_plots=True,
    sky_map_color_min=None,
    sky_map_color_max=None,
):
    """
    Assign new_value to all pixels in a skymap within pixel_radius of pixels with value <, >, or = find_value.

    Parameters
    ----------
    my_bundles : `dict` {`rubin_sim.maf.MetricBundles`}
        a dictionary for metricBundles.
    plot_handler :   `rubin_sim.maf.plots.plotHandler.PlotHandler`
    data_label : `str`
        description of the data, i.e. 'numGal'
    nside : `int`
        HEALpix resolution parameter. Default: 128
    find_value : `str`
        if related to mask, must be either 'masked' or 'unmasked'. otherwise, must be a number.
        Default: 'unmasked'
    relation : `str`
        must be '>','=','<'. Default: '='
    new_value : `str`
        if related to mask, must be either 'masked' or 'unmasked'; otherwise, must be a number.
        Default: 'masked'
    pixel_radius : `int`
        number of pixels to consider around a given pixel. Default: 6
    return_border_indices : `bool`
        set to True to return the array of indices of the pixels whose values/mask are changed. Default: False
    print_intermediate_info : `bool`
        set to False if do not want to print intermediate info. Default: True
    plot_intermediate_plots : `bool`
        set to False if do not want to plot intermediate plots. Default: True
    print_final_info : `bool`
        set to False if do not want to print final info, i.e. total pixels changed. Default: True
    plot_final_plots : `bool`
        set to False if do not want to plot the final plots. Default: True
    sky_map_color_min : float
        color_min label value for skymap plot_dict label. Default: None
    sky_map_color_max : float
        color_max label value for skymap plot_dict label. Default: None
    """
    # find pixels such that (pixelValue (relation) find_value) AND their neighbors dont have that (relation) find_value.
    # then assign new_value to all these pixels.
    # relation must be '>','=','<'
    # data indices are the pixels numbers ..
    # ------------------------------------------------------------------------
    # check whether need to mask anything at all
    if pixel_radius == 0:
        print("No masking/changing of the original data.")
        if return_border_indices:
            borders = {}
            for dither in my_bundles:
                borders[dither] = []

            return [my_bundles, borders]
        else:
            return my_bundles
    # ------------------------------------------------------------------------
    # make sure that relation is compatible with find_value
    if (find_value == "masked") | (find_value == "unmasked"):
        if relation != "=":
            print('ERROR: must have relation== "=" if find_value is related to mask.')
            print('Setting:  relation= "="\n')
            relation = "="
    # ------------------------------------------------------------------------
    # translate find_value into what has to be assigned
    find_value_to_consider = find_value
    if find_value.__contains__("mask"):
        if find_value == "masked":
            find_value_to_consider = True
        if find_value == "unmasked":
            find_value_to_consider = False

    # translate new_value into what has to be assigned
    new_value_to_assign = new_value
    if new_value.__contains__("mask"):
        if new_value == "masked":
            new_value_to_assign = True
        if new_value == "unmasked":
            new_value_to_assign = False

    # ------------------------------------------------------------------------
    borders = {}
    for dither in my_bundles:
        total_border_pixel = []
        if print_intermediate_info:
            print("Survey strategy: %s" % dither)

        # find the array to look at.
        if (find_value).__contains__("mask"):
            orig_array = my_bundles[dither].metricValues.mask.copy()  # mask array
        else:
            orig_array = my_bundles[dither].metricValues.data.copy()  # data array

        for r in range(0, pixel_radius):
            border_pixel = []
            temp_copy = copy.deepcopy(my_bundles)
            # ignore the pixels whose neighbors formed the border in previous run
            if r != 0:
                orig_array[total_border_pixel] = new_value_to_assign

            # find the pixels that satisfy the relation with find_value and whose neighbors dont
            for i in range(0, len(orig_array)):
                neighbors_pixels = hp.get_all_neighbours(nside, i)  # i is the pixel number
                for j in neighbors_pixels:
                    condition = None
                    if relation == "<":
                        condition = (orig_array[i] < find_value_to_consider) & (
                            orig_array[j] >= find_value_to_consider
                        )
                    if relation == "=":
                        condition = (orig_array[i] == find_value_to_consider) & (
                            orig_array[j] != find_value_to_consider
                        )
                    if relation == ">":
                        condition = (orig_array[i] > find_value_to_consider) & (
                            orig_array[j] <= find_value_to_consider
                        )
                    if condition == None:
                        raise ValueError("ERROR: invalid relation: %s" % relation)

                    if condition:
                        if j != -1:  # -1 entries correspond to inexistent neighbors
                            border_pixel.append(i)

            border_pixel = np.unique(border_pixel)
            total_border_pixel.extend(border_pixel)

            if print_intermediate_info:
                print("Border pixels from run %s: %s" % (r + 1, len(border_pixel)))
                print("Total pixels so far: %s\n" % len(total_border_pixel))

            # plot found pixels
            if plot_intermediate_plots:
                if new_value.__contains__("mask"):
                    temp_copy[dither].metricValues.mask[:] = new_value_to_assign
                    temp_copy[dither].metricValues.mask[total_border_pixel] = not (new_value_to_assign)
                    temp_copy[dither].metricValues.data[total_border_pixel] = -500
                    plot_dict = {
                        "xlabel": data_label,
                        "title": "%s: %s Round # %s" % (dither, data_label, r + 1),
                        "log_scale": False,
                        "labelsize": 9,
                        "color_min": -550,
                        "color_max": 550,
                        "cmap": cm.jet,
                    }
                else:
                    temp_copy[dither].metricValues.mask[:] = True
                    temp_copy[dither].metricValues.mask[total_border_pixel] = False
                    temp_copy[dither].metricValues.data[total_border_pixel] = new_value_to_assign
                    plot_dict = {
                        "xlabel": data_label,
                        "title": "%s %s Round # %s" % (dither, data_label, r + 1),
                        "log_scale": False,
                        "labelsize": 9,
                        "maxl": 500,
                        "cmap": cm.jet,
                    }
                temp_copy[dither].set_plot_dict(plot_dict)
                temp_copy[dither].set_plot_funcs([plots.HealpixSkyMap(), plots.HealpixPowerSpectrum()])
                temp_copy[dither].plot(plot_handler=plot_handler)
                plt.show()
            # save the found pixels with the appropriate key
            borders[dither] = total_border_pixel

    # ------------------------------------------------------------------------
    # change the original map/array now.
    for dither in my_bundles:
        total_border_pixel = borders[dither]

        if print_final_info:
            print("Survey strategy: %s" % dither)
            print("Total pixels changed: %s\n" % len(total_border_pixel))

        if new_value.__contains__("mask"):
            my_bundles[dither].metricValues.mask[total_border_pixel] = new_value_to_assign
        else:
            my_bundles[dither].metricValues.data[total_border_pixel] = new_value_to_assign

        if plot_final_plots:
            # skymap
            plot_dict = {
                "xlabel": data_label,
                "title": "%s: %s MaskedMap; pixel_radius: %s " % (dither, data_label, pixel_radius),
                "log_scale": False,
                "labelsize": 8,
                "color_min": sky_map_color_min,
                "color_max": sky_map_color_max,
                "cmap": cm.jet,
            }
            my_bundles[dither].set_plot_dict(plot_dict)
            my_bundles[dither].set_plot_funcs([plots.HealpixSkyMap()])
            my_bundles[dither].plot(plot_handler=plot_handler)
            # power spectrum
            plot_dict = {
                "xlabel": data_label,
                "title": "%s: %s MaskedMap; pixel_radius: %s " % (dither, data_label, pixel_radius),
                "log_scale": False,
                "labelsize": 12,
                "maxl": 500,
                "cmap": cm.jet,
            }
            my_bundles[dither].set_plot_dict(plot_dict)
            my_bundles[dither].set_plot_funcs([plots.HealpixPowerSpectrum()])
            my_bundles[dither].plot(plot_handler=plot_handler)
            plt.show()

    if return_border_indices:
        return [my_bundles, borders]
    else:
        return my_bundles



__all__ = ("ExgalM5",)

from rubin_sim.phot_utils import DustValues

from .base_metric import BaseMetric
from .simple_metrics import Coaddm5Metric


class ExgalM5(BaseMetric):
    """
    Calculate co-added five-sigma limiting depth after dust extinction.

    Uses phot_utils to calculate dust extinction.

    Parameters
    ----------
    m5_col : `str`, optional
        Column name for five sigma depth. Default 'fiveSigmaDepth'.
    unit : `str`, optional
        Label for units. Default 'mag'.
    """

    def __init__(
        self, m5_col="fiveSigmaDepth", metric_name="ExgalM5", units="mag", filter_col="filter", **kwargs
    ):
        # Set the name for the dust map to use. This is gathered into the MetricBundle.
        maps = ["DustMap"]
        self.m5_col = m5_col
        self.filter_col = filter_col
        super().__init__(
            col=[self.m5_col, self.filter_col], maps=maps, metric_name=metric_name, units=units, **kwargs
        )
        # Set the default wavelength limits for the lsst filters. These are approximately correct.
        dust_properties = DustValues()
        self.ax1 = dust_properties.ax1
        # We will call Coaddm5Metric to calculate the coadded depth. Set it up here.
        self.coaddm5_metric = Coaddm5Metric(m5_col=m5_col)

    def run(self, data_slice, slice_point):
        """
        Compute the co-added m5 depth and then apply dust extinction to that magnitude.
        """
        m5 = self.coaddm5_metric.run(data_slice)
        if m5 == self.coaddm5_metric.badval:
            return self.badval
        # Total dust extinction along this line of sight. Correct default A to this EBV value.
        a_x = self.ax1[data_slice[self.filter_col][0]] * slice_point["ebv"]
        return m5 - a_x



"""A group of metrics that work together to evaluate season characteristics (length, number, etc).
In addition, these supports the time delay metric calculation for strong lensing.
"""
__all__ = (
    "find_season_edges",
    "SeasonLengthMetric",
    "CampaignLengthMetric",
    "MeanCampaignFrequencyMetric",
    "TdcMetric",
)

import numpy as np
from rubin_scheduler.utils import calc_season

from rubin_sim.phot_utils import DustValues

from .base_metric import BaseMetric


def find_season_edges(seasons):
    """Given the seasons, return the indexes of each start/end of the season.

    Parameters
    ----------
    seasons: np.ndarray
        Seasons, such as calculated by calc_season.
        Note that seasons should be sorted!!

    Returns
    -------
    np.ndarray, np.ndarray
        The indexes of the first and last date in the season.
    """
    int_seasons = np.floor(seasons)
    # Get the unique seasons, so that we can separate each one
    season_list = np.unique(int_seasons)
    # Find the first and last observation of each season.
    first_of_season = np.searchsorted(int_seasons, season_list)
    last_of_season = np.searchsorted(int_seasons, season_list, side="right") - 1
    return first_of_season, last_of_season


class SeasonLengthMetric(BaseMetric):
    """
    Calculate the length of LSST seasons, in days.

    Parameters
    ----------
    min_exp_time: float, optional
        Minimum visit exposure time to count for a 'visit', in seconds. Default 20.
    reduce_func : function, optional
       Function that can operate on array-like structures. Typically numpy function.
       This reduces the season length in each season from 10 separate values to a single value.
       Default np.median.
    """

    def __init__(
        self,
        mjd_col="observationStartMJD",
        exp_time_col="visitExposureTime",
        min_exp_time=16,
        reduce_func=np.median,
        metric_name="SeasonLength",
        **kwargs,
    ):
        units = "days"
        self.mjd_col = mjd_col
        self.exp_time_col = exp_time_col
        self.min_exp_time = min_exp_time
        self.reduce_func = reduce_func
        super().__init__(
            col=[self.mjd_col, self.exp_time_col], units=units, metric_name=metric_name, **kwargs
        )

    def run(self, data_slice, slice_point):
        """Calculate the (reduceFunc) of the length of each season.
        Uses the slice_point RA/Dec to calculate the position in question, then uses the times of the visits
        to assign them into seasons (based on where the sun is relative to the slice_point RA).

        Parameters
        ----------
        data_slice : numpy.array
            Numpy structured array containing the data related to the visits provided by the slicer.
        slice_point : dict
            Dictionary containing information about the slice_point currently active in the slicer.

        Returns
        -------
        float
           The (reduceFunc) of the length of each season, in days.
        """
        # Order data Slice/times and exclude visits which are too short.
        long = np.where(data_slice[self.exp_time_col] > self.min_exp_time)
        if len(long[0]) == 0:
            return self.badval
        data = np.sort(data_slice[long], order=self.mjd_col)
        # SlicePoints ra/dec are always in radians - convert to degrees to calculate season
        seasons = calc_season(np.degrees(slice_point["ra"]), data[self.mjd_col])
        first_of_season, last_of_season = find_season_edges(seasons)
        seasonlengths = data[self.mjd_col][last_of_season] - data[self.mjd_col][first_of_season]
        result = self.reduce_func(seasonlengths)
        return result


class CampaignLengthMetric(BaseMetric):
    """Calculate the number of seasons (roughly, years) a pointing is observed for.
    This corresponds to the 'campaign length' for lensed quasar time delays.
    """

    def __init__(
        self, mjd_col="observationStartMJD", exp_time_col="visitExposureTime", min_exp_time=20, **kwargs
    ):
        units = ""
        self.exp_time_col = exp_time_col
        self.min_exp_time = min_exp_time
        self.mjd_col = mjd_col
        super().__init__(col=[self.mjd_col, self.exp_time_col], units=units, **kwargs)

    def run(self, data_slice, slice_point):
        # Order data Slice/times and exclude visits which are too short.
        long = np.where(data_slice[self.exp_time_col] > self.min_exp_time)
        if len(long[0]) == 0:
            return self.badval
        data = np.sort(data_slice[long], order=self.mjd_col)
        seasons = calc_season(np.degrees(slice_point["ra"]), data[self.mjd_col])
        int_seasons = np.floor(seasons)
        count = len(np.unique(int_seasons))
        return count


class MeanCampaignFrequencyMetric(BaseMetric):
    """Calculate the mean separation between nights, within a season - then the mean over the campaign.
    Calculate per season, to avoid any influence from season gaps.
    """

    def __init__(
        self,
        mjd_col="observationStartMJD",
        exp_time_col="visitExposureTime",
        min_exp_time=20,
        night_col="night",
        **kwargs,
    ):
        self.mjd_col = mjd_col
        self.exp_time_col = exp_time_col
        self.min_exp_time = min_exp_time
        self.night_col = night_col
        units = "nights"
        super().__init__(col=[self.mjd_col, self.exp_time_col, self.night_col], units=units, **kwargs)

    def run(self, data_slice, slice_point):
        # Order data Slice/times and exclude visits which are too short.
        long = np.where(data_slice[self.exp_time_col] > self.min_exp_time)
        if len(long[0]) == 0:
            return self.badval
        data = np.sort(data_slice[long], order=self.mjd_col)
        # SlicePoints ra/dec are always in radians - convert to degrees to calculate season
        seasons = calc_season(np.degrees(slice_point["ra"]), data[self.mjd_col])
        first_of_season, last_of_season = find_season_edges(seasons)
        season_means = np.zeros(len(first_of_season), float)
        for i, (first, last) in enumerate(zip(first_of_season, last_of_season)):
            if first < last:
                n = data[self.night_col][first : last + 1]
                delta_nights = np.diff(np.unique(n))
                if len(delta_nights) > 0:
                    season_means[i] = np.mean(delta_nights)
        return np.mean(season_means)


class TdcMetric(BaseMetric):
    """Calculate the Time Delay Challenge metric, as described in Liao et al 2015
    (https://arxiv.org/pdf/1409.1254.pdf).

    This combines the MeanCampaignFrequency/MeanNightSeparation, the SeasonLength, and the CampaignLength
    metrics above, but rewritten to calculate season information only once.

    cad_norm = in units of days
    sea_norm = in units of months
    camp_norm = in units of years

    This metric also adds a requirement to achieve limiting magnitudes after galactic dust extinction,
    in various bandpasses, in order to exclude visits which are not useful for detecting quasars
    (due to being short or having high sky brightness, etc.) and to reject regions with
    high galactic dust extinction.

    Parameters
    ----------
    mjd_col: str, optional
        Column name for mjd. Default observationStartMJD.
    night_col: str, optional
        Column name for night. Default night.
    filter_col: str, optional
        Column name for filter. Default filter.
    m5_col: str, optional
        Column name for five-sigma depth. Default fiveSigmaDepth.
    mag_cuts: dict, optional
        Dictionary with filtername:mag limit (after dust extinction). Default None in kwarg.
        Defaults set within metric: {'u': 22.7, 'g': 24.1, 'r': 23.7, 'i': 23.1, 'z': 22.2, 'y': 21.4}
    metricName: str, optional
        Metric Name. Default TDC.
    cad_norm: float, optional
        Cadence normalization constant, in units of days. Default 3.
    sea_norm: float, optional
        Season normalization constant, in units of months. Default 4.
    camp_norm: float, optional
        Campaign length normalization constant, in units of years. Default 5.
    badval: float, optional
        Return this value instead of the dictionary for bad points.

    Returns
    -------
    dictionary
        Dictionary of values for {'rate', 'precision', 'accuracy'} at this point in the sky.
    """

    def __init__(
        self,
        mjd_col="observationStartMJD",
        night_col="night",
        filter_col="filter",
        m5_col="fiveSigmaDepth",
        mag_cuts=None,
        metric_name="TDC",
        cad_norm=3.0,
        sea_norm=4.0,
        camp_norm=5.0,
        badval=-999,
        **kwargs,
    ):
        # Save the normalization values.
        self.cad_norm = cad_norm
        self.sea_norm = sea_norm
        self.camp_norm = camp_norm
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.night_col = night_col
        self.filter_col = filter_col
        if mag_cuts is None:
            self.mag_cuts = {
                "u": 22.7,
                "g": 24.1,
                "r": 23.7,
                "i": 23.1,
                "z": 22.2,
                "y": 21.4,
            }
        else:
            self.mag_cuts = mag_cuts
            if not isinstance(self.mag_cuts, dict):
                raise Exception("mag_cuts should be a dictionary")
        # Set up dust map requirement
        maps = ["DustMap"]
        # Set the default wavelength limits for the lsst filters. These are approximately correct.
        dust_properties = DustValues()
        self.ax1 = dust_properties.ax1
        super().__init__(
            col=[self.mjd_col, self.m5_col, self.night_col, self.filter_col],
            badval=badval,
            maps=maps,
            metric_name=metric_name,
            units="%s" % ("%"),
            **kwargs,
        )

    def run(self, data_slice, slice_point):
        # Calculate dust-extinction limiting magnitudes for each visit.
        filterlist = np.unique(data_slice[self.filter_col])
        m5_dust = np.zeros(len(data_slice), float)
        for f in filterlist:
            match = np.where(data_slice[self.filter_col] == f)
            a_x = self.ax1[f] * slice_point["ebv"]
            m5_dust[match] = data_slice[self.m5_col][match] - a_x
            m5_dust[match] = np.where(m5_dust[match] > self.mag_cuts[f], m5_dust[match], -999)
        idxs = np.where(m5_dust > -998)
        if len(idxs[0]) == 0:
            return self.badval
        data = np.sort(data_slice[idxs], order=self.mjd_col)
        # SlicePoints ra/dec are always in radians - convert to degrees to calculate season
        seasons = calc_season(np.degrees(slice_point["ra"]), data[self.mjd_col])
        int_seasons = np.floor(seasons)
        first_of_season, last_of_season = find_season_edges(seasons)
        # Campaign length
        camp = len(np.unique(int_seasons))
        # Season length
        seasonlengths = data[self.mjd_col][last_of_season] - data[self.mjd_col][first_of_season]
        sea = np.median(seasonlengths)
        # Convert to months
        sea = sea / 30.0
        # Campaign frequency / mean night separation
        season_means = np.zeros(len(first_of_season), float)
        for i, (first, last) in enumerate(zip(first_of_season, last_of_season)):
            n = data[self.night_col][first : last + 1]
            delta_nights = np.diff(np.unique(n))
            if len(delta_nights) > 0:
                season_means[i] = np.mean(delta_nights)
        cad = np.mean(season_means)
        # Evaluate precision and accuracy for TDC
        if sea == 0 or cad == 0 or camp == 0:
            return self.badval
        else:
            accuracy = 0.06 * (self.sea_norm / sea) * (self.camp_norm / camp) ** (1.1)
            precision = (
                4.0
                * (cad / self.cad_norm) ** (0.7)
                * (self.sea_norm / sea) ** (0.3)
                * (self.camp_norm / camp) ** (0.6)
            )
            rate = (
                30.0
                * (self.cad_norm / cad) ** (0.4)
                * (sea / self.sea_norm) ** (0.8)
                * (self.camp_norm / camp) ** (0.2)
            )
        return {
            "accuracy": accuracy,
            "precision": precision,
            "rate": rate,
            "cadence (days)": cad,
            "season (months)": sea,
            "campaign": camp,
        }

    def reduce_accuracy(self, metric_value):
        return metric_value["accuracy"]

    def reduce_precision(self, metric_value):
        return metric_value["precision"]

    def reduce_rate(self, metric_value):
        return metric_value["rate"]

    def reduce_cadence(self, metric_value):
        return metric_value["cadence (days)"]

    def reduce_season(self, metric_value):
        return metric_value["season (months)"]

    def reduce_campaign(self, metric_value):
        return metric_value["campaign"]



__all__ = ("ChebyFits",)

import os
import warnings

import numpy as np

from .chebyshev_utils import chebfit, make_cheb_matrix, make_cheb_matrix_only_x
from .ooephemerides import PyOrbEphemerides, get_oorb_data_dir
from .orbits import Orbits


def three_sixty_to_neg(ra):
    """Wrap discontiguous RA values into more-contiguous results."""
    if (ra.min() < 100) and (ra.max() > 270):
        ra = np.where(ra > 270, ra - 360, ra)
    return ra


class ChebyFits:
    """Generates chebyshev coefficients for a provided set of orbits.

    Calculates true ephemerides using PyEphemerides, then fits these positions with a constrained
    Chebyshev Polynomial, using the routines in chebyshevUtils.py.
    Many chebyshev polynomials are used to fit one moving object over a given timeperiod;
    typically, the length of each segment is typically about 2 days for MBAs.
    The start and end of each segment must match exactly, and the entire segments must
    fit into the total timespan an integer number of times. This is accomplished by setting n_decimal to
    the number of decimal places desired in the 'time' value. For faster moving objects, this number needs
    be greater to allow for smaller subdivisions. It's tempting to allow flexibility to the point of not
    enforcing this for non-database use; however, then the resulting ephemeris may have multiple values
    depending on which polynomial segment was used to calculate the ephemeris.
    The length of each chebyshev polynomial is related to the number of ephemeris positions used to fit that
    polynomial by ngran:
    length = timestep * ngran
    The length of each polynomial is adjusted so that the residuals in RA/Dec position
    are less than sky_tolerance - default = 2.5mas.
    The polynomial length (and the resulting residuals) is affected by ngran (i.e. timestep).

    Default values are based on Yusra AlSayaad's work.

    Parameters
    ----------
    orbits_obj : Orbits
        The orbits for which to fit chebyshev polynomial coefficients.
    t_start : float
        The starting point in time to fit coefficients. MJD.
    t_span : float
        The time span (starting at t_start) over which to fit coefficients. Days.
    time_scale : {'TAI', 'UTC', 'TT'}
        The timescale of the MJD time, t_start, and the time_scale that should be
        used with the chebyshev coefficients.
    obsCode : int, optional
        The observatory code of the location for which to generate ephemerides. Default 807 (CTIO).
    sky_tolerance : float, optional
        The desired tolerance in mas between ephemerides calculated by OpenOrb and fitted values.
        Default 2.5 mas.
    nCoeff_position : int, optional
        The number of Chebyshev coefficients to fit for the RA/Dec positions. Default 14.
    nCoeff_vmag : int, optional
        The number of Chebyshev coefficients to fit for the V magnitude values. Default 9.
    nCoeff_delta : int, optional
        The number of Chebyshev coefficients to fit for the distance between Earth/Object. Default 5.
    nCoeff_elongation : int, optional
        The number of Chebyshev coefficients to fit for the solar elongation. Default 5.
    ngran : int, optional
        The number of ephemeris points within each Chebyshev polynomial segment. Default 64.
    eph_file : str, optional
        The path to the JPL ephemeris file to use. Default is '$OORB_DATA/de405.dat'.
    n_decimal : int, optional
        The number of decimal places to allow in the segment length (and thus the times of the endpoints)
        can be limited to n_decimal places. Default 10.
        For LSST SIMS moving object database, this should be 13 decimal places for NEOs and 0 for all others.
    """

    def __init__(
        self,
        orbits_obj,
        t_start,
        t_span,
        time_scale="TAI",
        obscode=807,
        sky_tolerance=2.5,
        n_coeff_position=14,
        n_coeff_vmag=9,
        n_coeff_delta=5,
        n_coeff_elongation=6,
        ngran=64,
        eph_file=None,
        n_decimal=10,
    ):
        # Set up PyOrbEphemerides.
        if eph_file is None:
            self.eph_file = os.path.join(get_oorb_data_dir(), "de405.dat")
        else:
            self.eph_file = eph_file
        self.pyephems = PyOrbEphemerides(self.eph_file)
        # And then set orbits.
        self._set_orbits(orbits_obj)
        # Save input parameters.
        # We have to play some games with the start and end times, using Decimal,
        # in order to get the subdivision and times to match exactly, up to n_decimal places.
        self.n_decimal = int(n_decimal)
        self.t_start = round(t_start, self.n_decimal)
        self.t_span = round(t_span, self.n_decimal)
        self.t_end = round(self.t_start + self.t_span, self.n_decimal)
        # print('input times', self.t_start, self.t_span, self.t_end, orbits_obj.orbits.obj_id.as_matrix())
        if time_scale.upper() == "TAI":
            self.time_scale = "TAI"
        elif time_scale.upper() == "UTC":
            self.time_scale = "UTC"
        elif time_scale.upper() == "TT":
            self.time_scale = "TT"
        else:
            raise ValueError("Do not understand time_scale; use TAI, UTC or TT.")
        self.obscode = obscode
        self.sky_tolerance = sky_tolerance
        self.n_coeff = {}
        self.n_coeff["position"] = int(n_coeff_position)
        self.n_coeff["geo_dist"] = int(n_coeff_delta)
        self.n_coeff["vmag"] = int(n_coeff_vmag)
        self.n_coeff["elongation"] = int(n_coeff_elongation)
        self.ngran = int(ngran)
        # Precompute multipliers (we only do this once, instead of per segment).
        self._precompute_multipliers()
        # Initialize attributes to save the coefficients and residuals.
        self.coeffs = {
            "obj_id": [],
            "t_start": [],
            "t_end": [],
            "ra": [],
            "dec": [],
            "geo_dist": [],
            "vmag": [],
            "elongation": [],
        }
        self.resids = {
            "obj_id": [],
            "t_start": [],
            "t_end": [],
            "pos": [],
            "geo_dist": [],
            "vmag": [],
            "elongation": [],
        }
        self.failed = []

    def _set_orbits(self, orbits_obj):
        """Set the orbits, to be used to generate ephemerides.

        Parameters
        ----------
        orbits_obj : Orbits
           The orbits to use to generate ephemerides.
        """
        if not isinstance(orbits_obj, Orbits):
            raise ValueError("Need to provide an Orbits object.")
        self.orbits_obj = orbits_obj
        self.pyephems.set_orbits(self.orbits_obj)

    def _precompute_multipliers(self):
        """Calculate multipliers for Chebyshev fitting.

        Calculate these once, rather than for each segment.
        """
        # The nPoints are predetermined here, based on Yusra's earlier work.
        # The weight is based on Newhall, X. X. 1989, Celestial Mechanics, 45, p. 305-310
        self.multipliers = {}
        self.multipliers["position"] = make_cheb_matrix(self.ngran + 1, self.n_coeff["position"], weight=0.16)
        self.multipliers["vmag"] = make_cheb_matrix_only_x(self.ngran + 1, self.n_coeff["vmag"])
        self.multipliers["geo_dist"] = make_cheb_matrix_only_x(self.ngran + 1, self.n_coeff["geo_dist"])
        self.multipliers["elongation"] = make_cheb_matrix_only_x(self.ngran + 1, self.n_coeff["elongation"])

    def _length_to_timestep(self, length):
        """Convert chebyshev polynomial segment lengths to the corresponding timestep over the segment.

        Parameters
        ----------
        length : float
            The chebyshev polynomial segment length (nominally, days).

        Returns
        -------
        float
            The corresponding timestep, = length/ngran (nominally, days).
        """
        return length / self.ngran

    def make_all_times(self):
        """Using t_start and t_end, generate a numpy array containing times spaced at
        timestep = self.length/self.ngran.
        The expected use for this time array would be to generate ephemerides at each timestep.

        Returns
        -------
        np.ndarray
            Numpy array of times.
        """
        try:
            self.length
        except AttributeError:
            raise AttributeError("Need to set self.timestep first, using calcSegmentLength.")
        timestep = self._length_to_timestep(self.length)
        times = np.arange(self.t_start, self.t_end + timestep / 2, timestep)
        return times

    def generate_ephemerides(self, times, by_object=True):
        """Generate ephemerides using OpenOrb for all orbits.

        Parameters
        ----------
        times : np.ndarray
            The times to use for ephemeris generation.
        """
        return self.pyephems.generate_ephemerides(
            times,
            obscode=self.obscode,
            eph_mode="N",
            eph_type="basic",
            time_scale=self.time_scale,
            by_object=by_object,
        )

    def _round_length(self, length):
        """Modify length, to fit in an 'integer multiple' within the t_start/t_end,
        and to have the desired number of decimal values.

        Parameters
        ----------
        length : float
            The input length value to be rounded.

        Returns
        -------
        float
            The rounded length value.
        """
        length = round(length, self.n_decimal)
        length_in = length
        # Make length an integer value within the time interval, to last decimal place accuracy.
        counter = 0
        prev_int_factor = 0
        num_tolerance = 10.0 ** (-1 * (self.n_decimal - 1))
        while ((self.t_span % length) > num_tolerance) and (length > 0) and (counter < 20):
            int_factor = int(self.t_span / length) + 1  # round up / ceiling
            if int_factor == prev_int_factor:
                int_factor = prev_int_factor + 1
            prev_int_factor = int_factor
            length = round(self.t_span / int_factor, self.n_decimal)
            counter += 1
        if (self.t_span % length) > num_tolerance or (length <= 0):
            # Add this entire segment into the failed list.
            for obj_id in self.orbits_obj.orbits["obj_id"].as_matrix():
                self.failed.append((obj_id, self.t_start, self.t_end))
            raise ValueError(
                "Could not find a suitable length for the timespan (start %f, span %f), "
                "starting with length %s, ending with length value %f"
                % (self.t_start, self.t_span, str(length_in), length)
            )
        return length

    def _test_residuals(self, length, cutoff=99):
        """Calculate the position residual, for a test case.
        Convenience function to make calcSegmentLength easier to read.
        """
        # The pos_resid used will be the 'cutoff' percentile of all max residuals per object.
        max_pos_resids = np.zeros(len(self.orbits_obj), float)
        timestep = self._length_to_timestep(length)
        # Test for one segment near the start (would do at midpoint, but for long timespans
        # this is not efficient .. a point near the start should be fine).
        times = np.arange(self.t_start, self.t_start + length + timestep / 2, timestep)
        # We must regenerate ephemerides here, because the timestep is different each time.
        ephs = self.generate_ephemerides(times, by_object=True)
        # Look for the coefficients and residuals.
        for i, e in enumerate(ephs):
            coeff_ra, coeff_dec, max_pos_resids[i] = self._get_coeffs_position(e)
        # Find a representative value and return.
        pos_resid = np.percentile(max_pos_resids, cutoff)
        ratio = pos_resid / self.sky_tolerance
        return pos_resid, ratio

    def calc_segment_length(self, length=None):
        """Set the typical initial ephemeris timestep and segment length for all objects between t_start/t_end.

        Sets self.length.

        The segment length will fit into the time period between t_start/t_end an approximately integer
        multiple of times, and will only have a given number of decimal places.

        Parameters
        ----------
        length : float, optional
            If specified, this value for the length is used, instead of calculating it here.
        """
        # If length is specified, use it and do nothing else.
        if length is not None:
            length = self._round_length(length)
            pos_resid, ratio = self._test_residuals(length)
            if pos_resid > self.sky_tolerance:
                warnings.warn(
                    "Will set length and timestep, but this value of length "
                    "produces residuals (%f) > skyTolerance (%f)." % (pos_resid, self.sky_tolerance)
                )
            self.length = length
            return
        # Otherwise, calculate an appropriate length and timestep.
        # Give a guess at a very approximate segment length, given the skyTolerance,
        # purposefully trying to overestimate this value.
        # The actual behavior of the residuals is not linear with segment length.
        # There is a linear increase at low residuals < ~2 mas / segment length < 2 days
        # Then at around 2 days the residuals blow up, increasing rapidly to about 5000 mas
        #   (depending on orbit .. TNOs, for example, increase but only to about 300 mas,
        #    when the residuals resume ~linear growth out to 70 day segments if ngran=128)
        # Make an arbitrary cap on segment length at 60 days, (25000 mas) ~.5 arcminute accuracy.
        max_length = 60
        max_iterations = 50
        if self.sky_tolerance < 5:
            # This is the cap of the low-linearity regime, looping below will refine this value.
            length = 2.0
        elif self.sky_tolerance >= 5000:
            # Make a very rough guess.
            length = np.round((5000.0 / 20.0) * (self.sky_tolerance - 5000.0)) + 5.0
            length = np.min([max_length, int(length * 10) / 10.0])
        else:
            # Try to pick a length somewhere in the middle of the fast increase.
            length = 4.0
        # Tidy up some characteristics of "length":
        # make it fit an integer number of times into overall timespan.
        # and use a given number of decimal places (easier for database storage).
        length = self._round_length(length)
        # Check the resulting residuals.
        pos_resid, ratio = self._test_residuals(length)
        counter = 0
        # Now should be relatively close. Start to zero in using slope around the value.ngran
        while pos_resid > self.sky_tolerance and counter <= max_iterations and length > 0:
            length = length / 2
            length = self._round_length(length)
            pos_resid, ratio = self._test_residuals(length)
            counter += 1
            # print(counter, length, pos_resid, ratio)
        if counter > max_iterations or length <= 0:
            # Add this entire segment into the failed list.
            for obj_id in self.orbits_obj.orbits["obj_id"].as_matrix():
                self.failed.append((obj_id, self.t_start, self.t_end))
            error_message = "Could not find good segment length to meet skyTolerance %f" % (
                self.sky_tolerance
            )
            error_message += " milliarcseconds within %d iterations. " % (max_iterations)
            error_message += "Final residual was %f milli-arcseconds." % (pos_resid)
            raise ValueError(error_message)
        else:
            self.length = length

    def _get_coeffs_position(self, ephs):
        """Calculate coefficients for the ra/dec values of a single objects ephemerides.

        Parameters
        ----------
        times : np.ndarray
            The times of the ephemerides.
        ephs : np.ndarray
            The structured array returned by PyOrbEphemerides holding ephemeris values, for one object.

        Returns
        -------
        np.ndarray
            The ra coefficients
        np.ndarray
            The dec coefficients
        float
            The positional error residuals between fit and ephemeris values, in mas.
        """
        dradt_coord = ephs["dradt"] / np.cos(np.radians(ephs["dec"]))
        coeff_ra, resid_ra, rms_ra_resid, max_ra_resid = chebfit(
            ephs["time"],
            three_sixty_to_neg(ephs["ra"]),
            dxdt=dradt_coord,
            x_multiplier=self.multipliers["position"][0],
            dx_multiplier=self.multipliers["position"][1],
            n_poly=self.n_coeff["position"],
        )
        coeff_dec, resid_dec, rms_dec_resid, max_dec_resid = chebfit(
            ephs["time"],
            ephs["dec"],
            dxdt=ephs["ddecdt"],
            x_multiplier=self.multipliers["position"][0],
            dx_multiplier=self.multipliers["position"][1],
            n_poly=self.n_coeff["position"],
        )
        max_pos_resid = np.max(np.sqrt(resid_dec**2 + (resid_ra * np.cos(np.radians(ephs["dec"]))) ** 2))
        # Convert position residuals to mas.
        max_pos_resid *= 3600.0 * 1000.0
        return coeff_ra, coeff_dec, max_pos_resid

    def _get_coeffs_other(self, ephs):
        """Calculate coefficients for the ra/dec values of a single objects ephemerides.

        Parameters
        ----------
        ephs : np.ndarray
            The structured array returned by PyOrbEphemerides holding ephemeris values, for one object.

        Returns
        -------
        dict
            Dictionary containing the coefficients for each of 'geo_dist', 'vmag', 'elongation'
        dict
            Dictionary containing the max residual values for each of 'geo_dist', 'vmag', 'elongation'.
        """
        coeffs = {}
        max_resids = {}
        for key, ephValue in zip(("geo_dist", "vmag", "elongation"), ("geo_dist", "magV", "solarelon")):
            coeffs[key], resid, rms, max_resids[key] = chebfit(
                ephs["time"],
                ephs[ephValue],
                dxdt=None,
                x_multiplier=self.multipliers[key],
                dx_multiplier=None,
                n_poly=self.n_coeff[key],
            )
        return coeffs, max_resids

    def calc_segments(self):
        """Run the calculation of all segments over the entire time span."""
        # First calculate ephemerides for all objects, over entire time span.
        # For some objects, we will end up recalculating the ephemeride values, but most should be fine.
        times = self.make_all_times()
        ephs = self.generate_ephemerides(times)
        eps = self._length_to_timestep(self.length) / 4.0
        # Loop through each object to generate coefficients.
        for orbit_obj, e in zip(self.orbits_obj, ephs):
            t_segment_start = self.t_start
            # Cycle through all segments until we reach the end of the period we're fitting.
            while t_segment_start < (self.t_end - eps):
                # Identify the subset of times and ephemerides which are relevant for this segment
                # (at the default segment size).
                t_segment_end = round(t_segment_start + self.length, self.n_decimal)
                subset = np.where((times >= t_segment_start) & (times < t_segment_end + eps))
                self.calc_one_segment(orbit_obj, e[subset])
                t_segment_start = t_segment_end

    def calc_one_segment(self, orbit_obj, ephs):
        """Calculate the coefficients for a single Chebyshev segment, for a single object.

        Calculates the coefficients and residuals, and saves this information to self.coeffs,
        self.resids, and (if there are problems), self.failed.

        Parameters
        ----------
        orbit_obj : Orbits
            The single Orbits object we're fitting at the moment.
        ephs : np.ndarray
            The ephemerides we're fitting at the moment (for the single object / single segment).
        """
        obj_id = orbit_obj.orbits.obj_id.iloc[0]
        t_segment_start = ephs["time"][0]
        t_segment_end = ephs["time"][-1]
        coeff_ra, coeff_dec, max_pos_resid = self._get_coeffs_position(ephs)
        if max_pos_resid > self.sky_tolerance:
            # print('subdividing segments', orbit_obj.orbits.obj_id.iloc[0])
            self._subdivide_segment(orbit_obj, ephs)
        else:
            # print('working on ', orbit_obj.orbits.obj_id.iloc[0], 'at times', t_segment_start, t_segment_end)
            coeffs, max_resids = self._get_coeffs_other(ephs)
            fit_failed = False
            for k in max_resids:
                if np.isnan(max_resids[k]):
                    fit_failed = True
            if fit_failed:
                warnings.warn(
                    "Fit failed for orbit_obj %s for times between %f and %f"
                    % (obj_id, t_segment_start, t_segment_end)
                )
                self.failed.append((orbit_obj.orbits["obj_id"], t_segment_start, t_segment_end))
            else:
                # Consolidate items into the tracked coefficient values.
                self.coeffs["obj_id"].append(obj_id)
                self.coeffs["t_start"].append(t_segment_start)
                self.coeffs["t_end"].append(t_segment_end)
                self.coeffs["ra"].append(coeff_ra)
                self.coeffs["dec"].append(coeff_dec)
                self.coeffs["geo_dist"].append(coeffs["geo_dist"])
                self.coeffs["vmag"].append(coeffs["vmag"])
                self.coeffs["elongation"].append(coeffs["elongation"])
                # Consolidate items into the tracked residual values.
                self.resids["obj_id"].append(obj_id)
                self.resids["t_start"].append(t_segment_start)
                self.resids["t_end"].append(t_segment_end)
                self.resids["pos"].append(max_pos_resid)
                self.resids["geo_dist"].append(max_resids["geo_dist"])
                self.resids["vmag"].append(max_resids["geo_dist"])
                self.resids["elongation"].append(max_resids["elongation"])

    def _subdivide_segment(self, orbit_obj, ephs):
        """Subdivide a segment, then calculate the segment coefficients.

        Parameters
        ----------
        orbit_obj : Orbits
            The single Orbits object we're fitting at the moment.
        ephs : np.ndarray
            The ephemerides we're fitting at the moment (for the single object / single segment).
        """
        new_cheby = ChebyFits(
            orbit_obj,
            ephs["time"][0],
            (ephs["time"][-1] - ephs["time"][0]),
            time_scale=self.time_scale,
            obscode=self.obscode,
            sky_tolerance=self.sky_tolerance,
            n_coeff_position=self.n_coeff["position"],
            n_coeff_vmag=self.n_coeff["vmag"],
            n_coeff_delta=self.n_coeff["geo_dist"],
            n_coeff_elongation=self.n_coeff["elongation"],
            ngran=self.ngran,
            eph_file=self.eph_file,
            n_decimal=self.n_decimal,
        )
        try:
            new_cheby.calc_segment_length()
        except ValueError as ve:
            # Could not find a good segment length.
            warningmessage = "Objid %s, segment %f to %f " % (
                orbit_obj.orbits.obj_id.iloc[0],
                ephs["time"][0],
                ephs["time"][-1],
            )
            warningmessage += " - error: %s" % (ve)
            warnings.warn(warningmessage)
            self.failed += new_cheby.failed
            return
        new_cheby.calc_segments()
        # Add subdivided segment values into tracked values here.
        for k in self.coeffs:
            self.coeffs[k] += new_cheby.coeffs[k]
        for k in self.resids:
            self.resids[k] += new_cheby.resids[k]
        self.failed += new_cheby.failed

    def write(self, coeff_file, resid_file, failed_file, append=False):
        """Write coefficients, residuals and failed fits to disk.

        Parameters
        ----------
        coeff_file : str
            The filename for the coefficient values.
        resid_file : str
            The filename for the residual values.
        failed_file : str
            The filename to write the failed fit information (if failed objects exist).
        append : bool, optional
            Flag to append (or overwrite) the output files.
        """

        warnings.warn(
            "Writing cheby fit values may have cross-platform issues. Consider passing values directly"
        )

        if append:
            open_mode = "aw"
        else:
            open_mode = "w"
        # Write a header to the coefficients file, if writing to a new file:
        if (not append) or (not os.path.isfile(coeff_file)):
            header = "obj_id t_start t_end "
            header += " ".join(["ra_%d" % x for x in range(self.n_coeff["position"])]) + " "
            header += " ".join(["dec_%d" % x for x in range(self.n_coeff["position"])]) + " "
            header += " ".join(["geo_dist_%d" % x for x in range(self.n_coeff["geo_dist"])]) + " "
            header += " ".join(["vmag_%d" % x for x in range(self.n_coeff["vmag"])]) + " "
            header += " ".join(["elongation_%d" % x for x in range(self.n_coeff["elongation"])])
        else:
            header = None
        if (not append) or (not os.path.isfile(resid_file)):
            resid_header = "obj_id segNum t_start t_end length pos geo_dist vmag elong"
        else:
            resid_header = None
        timeformat = "%." + "%s" % self.n_decimal + "f"
        with open(coeff_file, open_mode) as f:
            if header is not None:
                print(header, file=f)
            for i, (obj_id, t_start, t_end, cRa, cDec, cDelta, cVmag, cE) in enumerate(
                zip(
                    self.coeffs["obj_id"],
                    self.coeffs["t_start"],
                    self.coeffs["t_end"],
                    self.coeffs["ra"],
                    self.coeffs["dec"],
                    self.coeffs["geo_dist"],
                    self.coeffs["vmag"],
                    self.coeffs["elongation"],
                )
            ):
                print(
                    "%s %s %s %s %s %s %s %s"
                    % (
                        obj_id,
                        timeformat % t_start,
                        timeformat % t_end,
                        " ".join("%.14e" % j for j in cRa),
                        " ".join("%.14e" % j for j in cDec),
                        " ".join("%.7e" % j for j in cDelta),
                        " ".join("%.7e" % j for j in cVmag),
                        " ".join("%.7e" % j for j in cE),
                    ),
                    file=f,
                )

        with open(resid_file, open_mode) as f:
            if resid_header is not None:
                print(resid_header, file=f)
            for i, (obj_id, t_start, t_end, rPos, rDelta, rVmag, rE) in enumerate(
                zip(
                    self.resids["obj_id"],
                    self.resids["t_start"],
                    self.resids["t_end"],
                    self.resids["pos"],
                    self.resids["geo_dist"],
                    self.resids["vmag"],
                    self.resids["elongation"],
                )
            ):
                print(
                    "%s %i %.14f %.14f %.14f %.14e %.14e %.14e %.14e"
                    % (
                        obj_id,
                        i + 1,
                        t_start,
                        t_end,
                        (t_end - t_start),
                        rPos,
                        rDelta,
                        rVmag,
                        rE,
                    ),
                    file=f,
                )

        if len(self.failed) > 0:
            with open(failed_file, open_mode) as f:
                for i, failed in enumerate(self.failed):
                    print(" ".join([str(x) for x in failed]), file=f)
